unit RVXMLBase;

interface

uses SysUtils, Classes;

const
  crlf = #13#10;
  tab = #9;

type
  TXMLValueType = (vtText, vtCDATA);
  PPChar = ^PChar;

  TXMLAttr = class (TObject)
  public
    Name, Value: string;
  end;

  TXMLAttrList = class (TList)
  protected
    function Get(Index: Integer): TXMLAttr;
    procedure Put(Index: Integer; const Value: TXMLAttr);
  public
    function AddAttr(Name, Value: string): TXMLAttr;
    procedure Clear; override;
    function FindAttrOfName(Name: string): TXMLAttr;
    property Items[Index: Integer]: TXMLAttr read Get write Put; default;
  end;

  TXMLTagList = class;
  TXMLTree = class;

  TXMLTag = class (TObject)
  public
    Name, Value: string;
    UniValue: WideString;
    Level: Integer;
    ValueType: TXMLValueType;
    Items: TXMLTagList;
    Attr: TXMLAttrList;
    Parent: TXMLTag;
    Root: TXMLTree;
    Unicode: Boolean;
    procedure Clear;
    constructor Create(Str: string; AParent: TXMLTag; ARoot: TXMLTree);
    destructor Destroy; override;
    procedure GetDocument(S: TStrings; Tabs: Boolean);
    function NextSibling: TXMLTag;
  end;

  TXMLTagList = class (TList)
  protected
    function Get(Index: Integer): TXMLTag;
    procedure Put(Index: Integer; const Value: TXMLTag);
  public
    function AddTag(Str: string; Parent: TXMLTag; Root: TXMLTree): TXMLTag;
    procedure Clear; override;
    function FindTagOfName(Name: string): TXMLTag;
    property Items[Index: Integer]: TXMLTag read Get write Put; default;
  end;

  TXMLTree = class (TXMLTag)
  private
    procedure Parse(S: string);
  public
    UnicodeTags: TStringList;
    Encoding: string;
    BaseFilePath: string;
    constructor Create(Enc: string);
    destructor Destroy; override;
    procedure ExtractBaseFilePath(FileName: string);
    procedure GetDocument(S: TStrings; Tabs: Boolean);
    procedure LoadFromFile(FileName: string);
    procedure LoadFromStream(Stream: TStream);
    procedure SaveToFile(FileName: string; Tabs: Boolean);
    procedure SaveToStream(Stream: TStream; Tabs: Boolean);
  end;

implementation

function AdvPos(const FindString, SourceString: string; StartPos: Integer): Integer;
asm
        PUSH    ESI
        PUSH    EDI
        PUSH    EBX
        PUSH    EDX
        TEST    EAX,EAX
        JE      @@qt
        TEST    EDX,EDX
        JE      @@qt0
        MOV     ESI,EAX
        MOV     EDI,EDX
        MOV     EAX,[EAX-4]
        MOV     EDX,[EDX-4]
        DEC     EAX
        SUB     EDX,EAX
        DEC     ECX
        SUB     EDX,ECX
        JNG     @@qt0
        MOV     EBX,EAX
        XCHG    EAX,EDX
        NOP
        ADD     EDI,ECX
        MOV     ECX,EAX
        MOV     AL,BYTE PTR [ESI]
@@lp1:  CMP     AL,BYTE PTR [EDI]
        JE      @@uu
@@fr:   INC     EDI
        DEC     ECX
        JNZ     @@lp1
@@qt0:  XOR     EAX,EAX
        JMP     @@qt
@@ms:   MOV     AL,BYTE PTR [ESI]
        MOV     EBX,EDX
        JMP     @@fr
@@uu:   TEST    EDX,EDX
        JE      @@fd
@@lp2:  MOV     AL,BYTE PTR [ESI+EBX]
        XOR     AL,BYTE PTR [EDI+EBX]
        JNE     @@ms
        DEC     EBX
        JNE     @@lp2
@@fd:   LEA     EAX,[EDI+1]
        SUB     EAX,[ESP]
@@qt:   POP     ECX
        POP     EBX
        POP     EDI
        POP     ESI
end;

function StringReplace(const Str, OldStr, NewStr: string; Flags: TReplaceFlags): string;
var
  P: Cardinal;
begin
  P := 1;
  Result := Str;
  while True do
  begin
    P := AdvPos(OldStr, Result, P);
    if P = 0 then Exit;
    Delete(Result, P, Length(OldStr));
    Insert(NewStr, Result, P);
    Inc(P, Length(NewStr));
  end;
end;

function CleanUnicode(Str: string): WideString;
var
  I, P, P2, L: Integer;
  S: string;
begin
  Result := '';
  L := Length(Str);
  P := 1;
  P2 := 0;
  repeat
    P := AdvPos('&#', Str, P);
    if P = 0 then
    begin
      Result := Result + Copy(Str, P2 + 1, L - P2);
      Exit;
    end else
      Result := Result + Copy(Str, P2 + 1, P - P2 - 1);

    P2 := AdvPos(';', Str, P);
    if P2 = 0 then Exit;
    S := Copy(Str, P + 2, P2 - P - 2);
    I := StrToIntDef(S, 0);
    Result := Result + WideChar(I);
    P := P2;
    Inc(P);
  until P > L;
end;

function UnCleanUnicode(Str: string): string;
var
  W: Word;
  I: Integer;
  C: array[0..1] of Char;
begin
  I := 1;
  Result := '';
  while I <= Length(Str) do
  begin
    C[0] := Str[I];
    C[1] := Str[I + 1];
    W := Word(C);
    if W > 255 then
      Result := Result + '&#' + IntToStr(W) + ';'
    else
      Result := Result + C[0];
    Inc(I, 2);
  end;
end;

function CleanUni(Str: string): string;
begin
  Result := StringReplace(Str, crlf, ' ', [rfReplaceAll]);
  Result := StringReplace(Result, tab, ' ', [rfReplaceAll]);
  Result := StringReplace(Result, '&lt;', '<', [rfReplaceAll]);
  Result := StringReplace(Result, '&gt;', '>', [rfReplaceAll]);
  Result := StringReplace(Result, '&apos;', '''', [rfReplaceAll]);
  Result := StringReplace(Result, '&amp;', '&', [rfReplaceAll]);
end;

function Clean(Str: string): string;
var
  P, P2, L, I, J: Integer;
  S: string;
begin
  Result := StringReplace(Str, crlf, ' ', [rfReplaceAll]);
  Result := StringReplace(Result, tab, ' ', [rfReplaceAll]);
  Result := StringReplace(Result, '&lt;', '<', [rfReplaceAll]);
  Result := StringReplace(Result, '&gt;', '>', [rfReplaceAll]);
  Result := StringReplace(Result, '&apos;', '''', [rfReplaceAll]);
  L := Length(Result);
  P := 1;
  J := 0;
  repeat
    P := AdvPos('&#', Result, P - J);
    if P = 0 then Break;
    P2 := AdvPos(';', Result, P);
    if P2 = 0 then Break;
    J := P2 - P;
    S := Copy(Result, P + 2, P2 - P - 2);
    I := StrToIntDef(S, 0);
    Delete(Result, P, P2 - P + 1);
    Insert(Chr(I), Result, P);
//    Result := Copy(Result, 1, P - 1) + Chr(I) + Copy(Result, P + 1, MaxInt);
    P := P2;
    Inc(P);
  until P > L; 
  Result := StringReplace(Result, '&amp;', '&', [rfReplaceAll]);
end;

function CleanSpec(Str: string): string;
begin
  Result := StringReplace(Str, '&quot;', '"', [rfReplaceAll]);
end;

function UnCleanSpec(Str: string): string;
var
  P: PChar;
  I, J: Integer;
begin
  P := PChar(Str);
  Result := '';
  I := 1;
  J := 1;
  while True do
  begin
    while Ord(P^) >= 32 do
    begin
      Inc(P);
      Inc(I);
    end;
    Result := Result + Copy(Str, J, I - J);
    if P^ = #0 then Exit;
{    if P^ = #0 then
    begin
      Result := Result + Copy(Str, J, I - J);
      Exit;
    end;
    Result := Result + Copy(Str, J - 1, I - J); }
    Result := Result + '&#' + IntToStr(Ord(P^)) + ';';
//    J := I + 1;
    Inc(I);
    J := I;
    Inc(P);
  end;
end;

function UnCleanUni(Str: string): string;
begin
  Result := StringReplace(Str, '<', '&lt;', [rfReplaceAll]);
  Result := StringReplace(Result, '>', '&gt;', [rfReplaceAll]);
end;

function UnClean(Str: string): string;
begin
  Result := StringReplace(Str, '"', '&quot;', [rfReplaceAll]);
  Result := UnCleanSpec(Result);
end;

function UnCleanText(Str: string): string;
begin
  Result := StringReplace(Str, '&', '&amp;', [rfReplaceAll]);
  Result := StringReplace(Result, '<', '&lt;', [rfReplaceAll]);
  Result := StringReplace(Result, '>', '&gt;', [rfReplaceAll]);
  Result := UnCleanSpec(Result);
end;

{TXMLAttrList}

function TXMLAttrList.AddAttr(Name, Value: string): TXMLAttr;
begin
  Result := TXMLAttr.Create;
  Result.Name := Name;
  Result.Value := Value;
  inherited Add(Result);
end;

procedure TXMLAttrList.Clear;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
    Items[I].Free;
  inherited;
end;

function TXMLAttrList.Get(Index: Integer): TXMLAttr;
begin
  Result := inherited Get(Index);
end;

function TXMLAttrList.FindAttrOfName(Name: string): TXMLAttr;
var
  I: Integer;
begin
  Result := nil;
  for I := 0 to Count - 1 do
    if LowerCase(Items[I].Name) = LowerCase(Name) then
    begin
      Result := Items[I];
      Exit;
    end;
end;

procedure TXMLAttrList.Put(Index: Integer; const Value: TXMLAttr);
begin
  inherited Put(Index, Value);
end;

{TXMLTag}

procedure TXMLTag.Clear;
begin
  Name := '';
  Value := '';
  ValueType := vtText;
  Attr.Clear;
  Items.Clear;
end;

constructor TXMLTag.Create(Str: string; AParent: TXMLTag; ARoot: TXMLTree);
var
  P, P2, L: Integer;
  AName, AValue: string;
begin
  inherited Create;
  Items := TXMLTagList.Create;
  Attr := TXMLAttrList.Create;
  Unicode := False;
  ValueType := vtText;
  Value := '';
  Root := ARoot;
  Parent := AParent;
  if Parent <> nil then
    Level := Parent.Level + 1
  else
    Level := -1;
  Name := Trim(Str);
  P := AdvPos(' ', Str, 1);
  if P = 0 then Exit;
  Name := Copy(Str, 1, P - 1);
  L := Length(Str);
  repeat
    P2 := AdvPos('=', Str, P);
    if P2 = 0 then Exit;
    AName := Trim(Copy(Str, P + 1, P2 - P - 1));
    P := AdvPos('"', Str, P2);
    if P = 0 then Exit;
    P2 := AdvPos('"', Str, P + 1);
    if P2 = 0 then Exit;
    AValue := CleanSpec(Clean(Copy(Str, P + 1, P2 - P - 1)));
    P := P2;
    Inc(P);
    Attr.AddAttr(AName, AValue);
  until P > L; 
end;

destructor TXMLTag.Destroy;
begin
  Items.Free;
  Attr.Free;
  Items := nil;
  Attr  := nil;
  inherited Destroy;
end;

procedure TXMLTag.GetDocument(S: TStrings; Tabs: Boolean);
var
  Spc, Str, Str2: string;
  I: Integer;
begin
  if Tabs then
    Spc := StringOfChar(tab, Level);

  Str := Spc + '<' + Name;

  Str2 := '';
  for I := 0 to Attr.Count - 1 do
    Str2 := Str2 + ' ' + Attr[I].Name + '="' + UnClean(Attr[I].Value) + '"';
  if Str2 <> '' then
    Str := Str + Str2;

  if (Value = '') and (Items.Count = 0) then
  begin
    Str := Str + ' />';
    S.Append(Str);
    Exit;
  end else
    Str := Str + '>';

  if Value <> '' then
  begin
    if Unicode then
      Str := Str + UnCleanUni(UnCleanUnicode(Value))
    else
      Str := Str + UnCleanText(Value);
  end;

  if Items.Count = 0 then
  begin
    Str := Str + '</' + Name + '>';
    S.Append(Str);
    Exit;
  end else
    S.Append(Str);

  for I := 0 to Items.Count - 1 do
    Items[I].GetDocument(S, Tabs);

  S.Append(Spc + '</' + Name + '>');
end;

function TXMLTag.NextSibling: TXMLTag;
var
  I: Integer;
begin
  Result := nil;
  if Parent = nil then Exit;
  I := Parent.Items.IndexOf(Self);
  if I = -1 then Exit;
  if I = Parent.Items.Count - 1 then Exit;
  Result := Parent.Items[I + 1];
end;

{TXMLTagList}

function TXMLTagList.AddTag(Str: string; Parent: TXMLTag; Root: TXMLTree): TXMLTag;
begin
  Result := TXMLTag.Create(Str, Parent, Root);
  inherited Add(Result);
end;

procedure TXMLTagList.Clear;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
    Items[I].Free;
  inherited;
end;

function TXMLTagList.Get(Index: Integer): TXMLTag;
begin
  Result := inherited Get(Index);
end;

function TXMLTagList.FindTagOfName(Name: string): TXMLTag;
var
  I: Integer;
begin
  Result := nil;
  for I := 0 to Count - 1 do
    if LowerCase(Items[I].Name) = LowerCase(Name) then
    begin
      Result := Items[I];
      Exit;
    end;
end;

procedure TXMLTagList.Put(Index: Integer; const Value: TXMLTag);
begin
  inherited Put(Index, Value);
end;

{TXMLTree}

constructor TXMLTree.Create(Enc: string);
begin
  inherited Create('', nil, nil);
  Level := -1;
  Encoding := Enc;
  UnicodeTags := TStringList.Create;
end;

destructor TXMLTree.Destroy;
begin
  UnicodeTags.Free;
  UnicodeTags := nil;
  inherited Destroy;
end;

procedure TXMLTree.ExtractBaseFilePath(FileName: string);
begin
  BaseFilePath := ExtractFilePath(ExpandUNCFileName(FileName));
end;

procedure TXMLTree.GetDocument(S: TStrings; Tabs: Boolean);
begin
  S.Clear;
  if Items.Count > 0 then
    Items[0].GetDocument(S, Tabs);
end;

procedure TXMLTree.LoadFromFile(FileName: string);
var
  Stream: TFileStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TXMLTree.LoadFromStream(Stream: TStream);
var
  S: string;
  P: PPChar;
begin
  SetLength(S, Stream.Size - Stream.Position);
  P := PPChar(@S);
  Stream.Read(P^^, Stream.Size - Stream.Position);
  Parse(S);
end;

procedure TXMLTree.Parse(S: string);
var
  Tag, PITag: TXMLTag;
  P, P2, L: Integer;
  T: string;
  PIAttr: TXMLAttr;
begin
  Items.Clear;
  Attr.Clear;
  Tag := Self;
  L := Length(S);
  p := 1;
  repeat
    P2 := AdvPos('<', S, P);
    if P2 <> 0 then
    begin
      T := Copy(S, P, P2 - P);
      if T <> '' then
      begin
        if UnicodeTags.IndexOf(Tag.Name) = -1 then
          Tag.Value := Clean(T)
        else
          Tag.UniValue := CleanUnicode(CleanUni(T));
      end;
      P := P2;
      if UpperCase(Copy(S, P, 9)) = '<![CDATA[' then
      begin
        P2 := AdvPos(']]>', S, P);
        T := Copy(S, P + 9, P2 - P - 9);
        Tag.Value := T;
        Tag.ValueType := vtCDATA;
        P := P2 + 2;
      end else if Copy(S, P, 4) = '<!--' then
      begin
        P2 := AdvPos('-->', S, P);
        if P2 = 0 then Exit;
        P := P2 + 2;
      end else if Copy(S, P, 2) = '<?' then
      begin
        P2 := AdvPos('?>', S, P);
        if P2 = 0 then Exit;
        T := Copy(S, P + 2, P2 - P - 2);
        P := P2 + 1;
        PITag := TXMLTag.Create(T, nil, nil);
        try
          if LowerCase(PITag.Name) = 'xml' then
          begin
            PIAttr := PITag.Attr.FindAttrOfName('encoding');
            if PIAttr <> nil then
              Encoding := PIAttr.Value;
          end;
        finally
          PITag.Free;
        end;
      end else
      begin
        P2 := AdvPos('>', S, P);
        if P2 = 0 then Exit;
        T := Trim(Copy(S, P + 1, P2 - P - 1));
        P := P2;
        if T[1] = '/' then
        begin
          Tag := Tag.Parent;
          if Tag = nil then Exit;
        end else if T[Length(T)] = '/' then
        begin
          Delete(T, Length(T), 1);
          Tag.Items.AddTag(T, Tag, Self);
        end else
          Tag := Tag.Items.AddTag(T, Tag, Self);
      end;
    end else
    begin
      T := Copy(S, P, Length(S) - P);
      if T <> '' then
        Tag.Value := Clean(T);
      Exit;
    end;
    Inc(P);
  until P > L;
end;

procedure TXMLTree.SaveToFile(FileName: string; Tabs: Boolean);
var
  Stream: TFileStream;
begin
  Stream := TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(Stream, Tabs);
  finally
    Stream.Free;
  end;
end;

procedure TXMLTree.SaveToStream(Stream: TStream; Tabs: Boolean);
var
  S: string;
  StrList: TStringList;
begin
  S := '<?xml version="1.0" encoding="' + Encoding + '"?>' + crlf;
  Stream.Write(PChar(S)^, Length(S));
  StrList := TStringList.Create;
  GetDocument(StrList, Tabs);
  StrList.SaveToStream(Stream);
  StrList.Free;
end;

end.
