unit RVXMLRoutines;

interface

{$I RV_Defs.inc}

uses SysUtils, Classes, Windows, RVXMLBase, Graphics, RichView, RVStyle, CRVData,
  RVScroll, RVTable, RVFMisc, Controls, RVXMLMisc, RVEdit, RVItem, RVClasses,
  PtblRV, RVMarker, RVFuncs;

type

  TControlClass = class of TControl;

  TRVXMLUnknownTagEvent = procedure (Node: TXMLTag; RVData: TCustomRVData; ItemNo: Integer) of object;

  TRVXMLLoadControlEvent = function (S: string): TControl of object;

  TRVXMLUnknownItemEvent = procedure (RVData: TCustomRVData; ItemNo: Integer; Item: TCustomRVItemInfo;
    Node: TXMLTag) of object;

  TRVXMLSaveControlEvent = function (Control: TControl): string of object;

  TStyleLoadingMode = (slmIgnore, slmReplace, slmMerge, slmMap);

  TRichViewXMLStyle = record
    StyleNo, ParaNo: Integer;
    VAlign: TRVVAlign;
    BreakColor: TColor;
    BreakWidth: Byte;
  end;

  PRichViewXMLStyle = ^TRichViewXMLStyle;

const
  DefaultXMLStyle: TRichViewXMLStyle = (StyleNo: 0; ParaNo: 0; VAlign: rvvaBaseLine;
    BreakColor: clNone; BreakWidth: 1);

function RVLoadFromXML(RichView: TCustomRichView; RVPrint: TRVPrint;
  LoadPrint: Boolean;
  Node: TXMLTag; var InsertPosition: Integer;
  LoadBackground: Boolean; StyleLoadingMode: TStyleLoadingMode;
  OnUnknownTag: TRVXMLUnknownTagEvent;
  OnLoadControl: TRVXMLLoadControlEvent;
  EditMode: Boolean; var FullReformat: Boolean;
  var NonFirstItemsAdded: Integer;
  LoadTagMode: Boolean): Boolean;

procedure RVSaveToXML(RichView: TCustomRichView; Node: TXMLTag;
  RVPrint: TRVPrint; SavePrint: Boolean;
  Style: TRichViewXMLStyle; SaveStyles: Boolean;
  SaveBackground: Boolean; SaveImgToFiles: Boolean;
  SelStart, SelOffset, SelEnd, SelEndOffset: Integer;
  OnUnknownItem: TRVXMLUnknownItemEvent;
  OnSaveControl: TRVXMLSaveControlEvent;
  UseStyleNames: Boolean; SaveStyleNames: Boolean);

implementation

uses ImgList;

{-----------------------------LoadRVDataFromXML-----------------------------}

function LoadRVDataFromXML(RVData: TCustomRVData; Node: TXMLTag;
  Bullets: TImgListList; Default: PRichViewXMLStyle;
  var InsertPosition: Integer;
  PaletteMode: TRVPaletteAction; RVPalette: HPalette;
  RVLogPalette: PLogPalette; PrevCP: TRVCPInfo;
  StyleMap, ParaMap, ListMap: TRVIntegerList;
  OnUnknownTag: TRVXMLUnknownTagEvent;
  OnLoadControl: TRVXMLLoadControlEvent;
  RVStyle: TRVStyle;
  EditMode: Boolean;
  var FirstTime, FullReformat: Boolean;
  var NonFirstItemsAdded: Integer;
  TagsArePChars: Boolean): Boolean;

var Position, NewListNo: Integer;

  function FindNamedStyleNo(const Name: string): Integer;
  var
    I: Integer;
  begin
    for I := 0 to RVStyle.TextStyles.Count - 1 do
      if LowerCase(RVStyle.TextStyles[I].StyleName) = LowerCase(Name) then
      begin
        Result := I;
        Exit;
      end;

    Result := -1;
  end;

  function FindNamedParaNo(const Name: string): Integer;
  var
    I: Integer;
  begin
    for I := 0 to RVStyle.ParaStyles.Count - 1 do
      if LowerCase(RVStyle.ParaStyles[I].StyleName) = LowerCase(Name) then
      begin
        Result := I;
        Exit;
      end;

    Result := -1;
  end;

  function FindNamedListNo(const Name: string): Integer;
  var
    I: Integer;
  begin
    for I := 0 to RVStyle.ListStyles.Count - 1 do
      if LowerCase(RVStyle.ListStyles[I].StyleName) = LowerCase(Name) then
      begin
        Result := I;
        Exit;
      end;

    Result := -1;
  end;

  function InsertItem(var Text: string; var Item: TCustomRVItemInfo): Boolean;
  begin
    Item.BeforeLoading(rvlfOther);
    if (RVData.Items.Count = 0) or (Position = 0) then
      Item.SameAsPrev := False
    else if RVData.GetItem(Position - 1).GetBoolValue(rvbpFullWidth) then
      Item.SameAsPrev := False
    else if Item.SameAsPrev then
      Item.ParaNo := RVData.GetItem(Position - 1).ParaNo;

    if (Item.ParaNo > RVStyle.ParaStyles.Count) or (Item.ParaNo < 0) then
      Item.ParaNo := Default^.ParaNo;

    Item.BR := Item.BR and not Item.GetBoolValue(rvbpFullWidth);

    Item.UpdatePaletteInfo(PaletteMode, False, RVPalette, RVLogPalette);
    if FirstTime then
    begin
      Result := RVData.InsertFirstRVFItem(Position, Text, Item, EditMode, FullReformat, NewListNo);
      if not Result then Exit;
      if Item<>nil then begin
        Item.AfterLoading(rvlfOther);
        Inc(Position);
        InsertPosition := Position-1;
        FirstTime := False;
      end;
    end else begin
      Item.Inserting(RVData, Text, False);
      RVData.Items.InsertObject(Position, Text, Item);
      Item.Inserted(RVData, Position);
      if not (RVData is TRVTableCellData) then
        Item.AfterLoading(rvlfOther);
      {$IFNDEF RVDONOTUSELISTS}
      RVData.AddMarkerInList(Position);
      {$ENDIF}
      Inc(Position);
      Inc(NonFirstItemsAdded);
      Result := True;
    end;
  end;

  procedure ReadText(T: TXMLTag);
  var
    I: Integer;
    S: string;
    Item: TRVTextItemInfo;
  begin
    Item := TRVTextItemInfo.Create(RVData);
    Item.StyleNo := Default^.StyleNo;
    Item.ParaNo := Default^.ParaNo;
    for I := 0 to T.Attr.Count - 1 do
    begin
      S := LowerCase(T.Attr[I].Name);

      if (S = 'tag') and TagsArePChars then Item.Tag := Integer(StrNew(PChar(T.Attr[I].Value)))

      else if (S = 'tagnumber') and (not TagsArePChars) then Item.Tag := StrToIntDef(T.Attr[I].Value, Item.Tag)

      else if S = 'textstyleno' then Item.StyleNo := StrToIntDef(T.Attr[I].Value, Item.StyleNo)

      else if S = 'parastyleno' then Item.ParaNo := StrToIntDef(T.Attr[I].Value, Item.ParaNo)

      else if S = 'textstyle' then Item.StyleNo := FindNamedStyleNo(T.Attr[I].Value)

      else if S = 'parastyle' then Item.ParaNo := FindNamedParaNo(T.Attr[I].Value)

      else if S = 'br' then
      begin
        if T.Attr[I].Value = '0' then
          Item.SameAsPrev := True
        else if T.Attr[I].Value = '1' then
          Item.SameAsPrev := False;
      end

      else if S = 'newpara' then
      begin
        if T.Attr[I].Value = '1' then
          Item.BR := False
        else if T.Attr[I].Value = '0' then
          Item.BR := True;
      end;

    end;

    if Item.StyleNo = -1 then
      Item.StyleNo := Default^.StyleNo;
    if Item.ParaNo = -1 then
      Item.ParaNo := Default^.ParaNo;

    Item.StyleNo := StyleMap[Item.StyleNo];
    Item.ParaNo := ParaMap[Item.ParaNo];

    InsertItem(T.Value, TCustomRVItemInfo(Item));
  end;

  procedure ReadUnicodeText(T: TXMLTag);
  var
    I: Integer;
    Item: TRVTextItemInfo;
    S, Str: string;
  begin
    Item := TRVTextItemInfo.Create(RVData);
    Item.StyleNo := Default^.StyleNo;
    Item.ParaNo := Default^.ParaNo;
    for I := 0 to T.Attr.Count - 1 do
    begin
      S := LowerCase(T.Attr[I].Name);

      if (S = 'tag') and TagsArePChars then Item.Tag := Integer(StrNew(PChar(T.Attr[I].Value)))

      else if (S = 'tagnumber') and (not TagsArePChars) then Item.Tag := StrToIntDef(T.Attr[I].Value, Item.Tag)

      else if S = 'textstyleno' then Item.StyleNo := StrToIntDef(T.Attr[I].Value, Item.StyleNo)

      else if S = 'parastyleno' then Item.ParaNo := StrToIntDef(T.Attr[I].Value, Item.ParaNo)

      else if S = 'textstyle' then Item.StyleNo := FindNamedStyleNo(T.Attr[I].Value)

      else if S = 'parastyle' then Item.ParaNo := FindNamedParaNo(T.Attr[I].Value)

      else if S = 'br' then
      begin
        if T.Attr[I].Value = '0' then
          Item.SameAsPrev := True
        else if T.Attr[I].Value = '1' then
          Item.SameAsPrev := False;
      end

      else if S = 'newpara' then
      begin
        if T.Attr[I].Value = '1' then
          Item.BR := False
        else if T.Attr[I].Value = '0' then
          Item.BR := True;
      end;

    end;
    SetLength(Str, Length(T.UniValue) * 2);
    Move(PWideChar(T.UniValue)^, PChar(Str)^, Length(Str));
    Include(Item.ItemOptions, rvioUnicode);

    if Item.StyleNo = -1 then
      Item.StyleNo := Default^.StyleNo;
    if Item.ParaNo = -1 then
      Item.ParaNo := Default^.ParaNo;

    Item.StyleNo := StyleMap[Item.StyleNo];
    Item.ParaNo := ParaMap[Item.ParaNo];

    InsertItem(Str, TCustomRVItemInfo(Item));
  end;

  procedure ReadCheck(T: TXMLTag);
  var
    I: Integer;
    Tag: Integer;
    Vis: Boolean;
    S: string;
  begin
    Tag := 0;
    Vis := True;
    for I := 0 to T.Attr.Count - 1 do
    begin

      S := LowerCase(T.Attr[I].Name);

      if (S = 'tag') and TagsArePChars then Tag := Integer(StrNew(PChar(T.Attr[I].Value)))

      else if (S = 'tagnumber') and (not TagsArePChars) then Tag := StrToIntDef(T.Attr[I].Value, Tag)

      else if S = 'vis' then
      begin
        if T.Attr[I].Value = '0' then
          Vis := False
        else if T.Attr[I].Value = '1' then
          Vis := True;
      end;

    end;
    RVData.SetCheckpointInfo(Position - 1, Tag, T.Value, Vis);
  end;

  function ReadStandardItemProperties(const AttrName: String; Attr: TXMLAttr;
    Item: TCustomRVItemInfo): Boolean;
  begin
    Result := True;
    if (AttrName = 'tag') and TagsArePChars then
      Item.Tag := Integer(StrNew(PChar(Attr.Value)))
    else if (AttrName = 'tagnumber') and (not TagsArePChars) then
      Item.Tag := StrToIntDef(Attr.Value, Item.Tag)
    else if AttrName = 'parastyleno' then
      Item.ParaNo := StrToIntDef(Attr.Value, Item.ParaNo)
    else if AttrName = 'parastyle' then
      Item.ParaNo := FindNamedParaNo(Attr.Value)
    {$IFNDEF RVDONOTUSEITEMHINTS}
    else if AttrName = 'hint' then
      Item.Hint := Attr.Value
    {$ENDIF}
    else if AttrName = 'br' then
      begin
        if Attr.Value = '0' then
          Item.SameAsPrev := True
        else if Attr.Value = '1' then
          Item.SameAsPrev := False;
      end
    else if AttrName = 'newpara' then
      begin
        if Attr.Value = '1' then
          Item.BR := False
        else if Attr.Value = '0' then
          Item.BR := True;
      end
    else
      Result := False;
  end;

  function ReadNonTextItemProperties(const AttrName: String; Attr: TXMLAttr;
    Item: TRVNonTextItemInfo): Boolean;
  begin
    Result := ReadStandardItemProperties(AttrName, Attr, Item);
    if not Result then
      if AttrName = 'deleteprotect' then
      begin
        if Attr.Value = '1' then
          Item.DeleteProtect := True
        else if Attr.Value = '0' then
          Item.DeleteProtect := False;
      end;
  end;

  function ReadRectItemProperties(const AttrName: String; Attr: TXMLAttr;
    Item: TRVRectItemInfo): Boolean;
  begin
    Result := ReadNonTextItemProperties(AttrName, Attr, Item);
    if Result then
      exit;
    Result := True;
    if AttrName = 'valign' then
      begin
        if LowerCase(Attr.Value) = 'baseline' then
          Item.VAlign := rvvaBaseLine
        else if LowerCase(Attr.Value) = 'middle' then
          Item.VAlign := rvvaMiddle;
      end
    else if AttrName = 'vshift' then
      Item.VShift := StrToIntDef(Attr.Value, Item.VShift)
    else if AttrName = 'vshiftabs' then
      begin
        if Attr.Value = '1' then
          Item.VShiftAbs := True
        else if Attr.Value = '0' then
          Item.VShiftAbs := False;
      end
    else if AttrName = 'spacing' then
      Item.Spacing := StrToIntDef(Attr.Value, 1)
    else
      Result := False;
  end;

  procedure ReadImage(T: TXMLTag);
  var
    I: Integer;
    S: string;
    Stream: TMemoryStream;
    Pic: TPicture;
    Item: TRVGraphicItemInfo;
    Transparent: Boolean;
    TransparentMode: TTransparentMode;
    TransparentColor: TColor;
  begin
    Transparent := False;
    TransparentMode := tmAuto;
    TransparentColor := 0;

    Item := TRVGraphicItemInfo.Create(RVData);
    Item.StyleNo := rvsPicture;
    Item.ParaNo := Default^.ParaNo;
    Item.VAlign := Default^.VAlign;
    Item.Image := nil;
    Stream := nil;
    Pic := TPicture.Create;
    try
      for I := 0 to T.Attr.Count - 1 do
      begin
        S := LowerCase(T.Attr[I].Name);

        if ReadRectItemProperties(S, T.Attr[I], Item) then
          continue;

        if S = 'class' then
          Item.Image := RV_CreateGraphics(TGraphicClass(GetClass(T.Attr[I].Value)))

        else if S = 'img' then
        begin
          Stream := TMemoryStream.Create;
          RVFTextString2Stream(T.Attr[I].Value, Stream);
        end

        else if S = 'imgsource' then
        begin
          Pic.LoadFromFile(T.Root.BaseFilePath + T.Attr[I].Value);
          Item.Image := RV_CreateGraphics(TGraphicClass(Pic.Graphic.ClassType));
          Item.Image.Assign(Pic.Graphic);
        end

        else if S = 'width' then
          Item.ImageWidth := StrToIntDef(T.Attr[I].Value, Item.ImageWidth)

        else if S = 'height' then
          Item.ImageHeight := StrToIntDef(T.Attr[I].Value, Item.ImageHeight)

        else if S = 'transparent' then
        begin
          if T.Attr[I].Value = '1' then
            Transparent := True
          else if T.Attr[I].Value = '0' then
            Transparent := False;
        end

        else if S = 'transparentmode' then
        begin
          if LowerCase(T.Attr[I].Value) = 'auto' then
            TransparentMode := tmAuto
          else if LowerCase(T.Attr[I].Value) = 'fixed' then
            TransparentMode := tmFixed;
        end

        else if S = 'transparentcolor' then
          TransparentColor := StringToColor(T.Attr[I].Value)

        else if S = 'alt' then
          Item.Alt := T.Attr[I].Value

        else if S = 'imagefilename' then
          Item.ImageFileName := T.Attr[I].Value

        else if S = 'minheighonpage' then
          Item.MinHeightOnPage := StrToIntDef(T.Attr[I].Value,0)

        else if S = 'htmlimagesize' then
        begin
          if T.Attr[I].Value = '1' then
            Item.NoHTMLImageSize := False
          else if T.Attr[I].Value = '0' then
            Item.NoHTMLImageSize := True;
        end

      end;

      if Item.ParaNo = -1 then
        Item.ParaNo := Default^.ParaNo;

      Item.ParaNo := ParaMap[Item.ParaNo];

      if (Stream <> nil) and (Item.Image <> nil) then
      begin
        Stream.Position := 0;
        try
          Item.Image.LoadFromStream(Stream);
        finally
          Stream.Free;
        end;
        if Item.Image is TBitmap then
        begin
          TBitmap(Item.Image).Transparent := Transparent;
          TBitmap(Item.Image).TransparentMode := TransparentMode;
          TBitmap(Item.Image).TransparentColor := TransparentColor;
        end;
        InsertItem(T.Value, TCustomRVItemInfo(Item));
      end else if Item.Image = Pic.Graphic then
      begin
        if Item.Image is TBitmap then
        begin
          TBitmap(Item.Image).Transparent := Transparent;
          TBitmap(Item.Image).TransparentMode := TransparentMode;
          TBitmap(Item.Image).TransparentColor := TransparentColor;
        end;
        InsertItem(T.Value, TCustomRVItemInfo(Item));
      end;
    finally
      Pic.Free;
    end;
  end;

  procedure ReadHotImage(T: TXMLTag);
  var
    I: Integer;
    S: string;
    Stream: TMemoryStream;
    Pic: TPicture;
    Item: TRVHotGraphicItemInfo;
    Transparent: Boolean;
    TransparentMode: TTransparentMode;
    TransparentColor: TColor;
  begin
    Transparent := False;
    TransparentMode := tmAuto;
    TransparentColor := 0;

    Item := TRVHotGraphicItemInfo.Create(RVData);
    Item.StyleNo := rvsHotPicture;
    Item.ParaNo := Default^.ParaNo;
    Item.VAlign := Default^.VAlign;
    Item.Image := nil;
    Stream := nil;
    Pic := TPicture.Create;
    try
      for I := 0 to T.Attr.Count - 1 do
      begin
        S := LowerCase(T.Attr[I].Name);

        if ReadRectItemProperties(S, T.Attr[I], Item) then
          continue;

        if S = 'class' then
          Item.Image := RV_CreateGraphics(TGraphicClass(GetClass(T.Attr[I].Value)))

        else if S = 'img' then
        begin
          Stream := TMemoryStream.Create;
          RVFTextString2Stream(T.Attr[I].Value, Stream);
        end

        else if S = 'imgsource' then
        begin
          Pic.LoadFromFile(T.Root.BaseFilePath + T.Attr[I].Value);
          Item.Image := RV_CreateGraphics(TGraphicClass(Pic.Graphic.ClassType));
          Item.Image.Assign(Pic.Graphic);
        end

        else if S = 'width' then
          Item.ImageWidth := StrToIntDef(T.Attr[I].Value, Item.ImageWidth)

        else if S = 'height' then
          Item.ImageHeight := StrToIntDef(T.Attr[I].Value, Item.ImageHeight)

        else if S = 'transparent' then
        begin
          if T.Attr[I].Value = '1' then
            Transparent := True
          else if T.Attr[I].Value = '0' then
            Transparent := False;
        end

        else if S = 'transparentmode' then
        begin
          if LowerCase(T.Attr[I].Value) = 'auto' then
            TransparentMode := tmAuto
          else if LowerCase(T.Attr[I].Value) = 'fixed' then
            TransparentMode := tmFixed;
        end

        else if S = 'transparentcolor' then
          TransparentColor := StringToColor(T.Attr[I].Value)

        else if S = 'alt' then
          Item.Alt := T.Attr[I].Value

        else if S = 'imagefilename' then
          Item.ImageFileName := T.Attr[I].Value

        else if S = 'minheighonpage' then
          Item.MinHeightOnPage := StrToIntDef(T.Attr[I].Value,0)

        else if S = 'htmlimagesize' then
        begin
          if T.Attr[I].Value = '1' then
            Item.NoHTMLImageSize := False
          else if T.Attr[I].Value = '0' then
            Item.NoHTMLImageSize := True;
        end


      end;

      if Item.ParaNo = -1 then
        Item.ParaNo := Default^.ParaNo;

      Item.ParaNo := ParaMap[Item.ParaNo];

      if (Stream <> nil) and (Item.Image <> nil) then
      begin
        Stream.Position := 0;
        try
          Item.Image.LoadFromStream(Stream);
        finally
          Stream.Free;
        end;
        if Item.Image is TBitmap then
        begin
          TBitmap(Item.Image).Transparent := Transparent;
          TBitmap(Item.Image).TransparentMode := TransparentMode;
          TBitmap(Item.Image).TransparentColor := TransparentColor;
        end;
        InsertItem(T.Value, TCustomRVItemInfo(Item));
      end else if Item.Image = Pic.Graphic then
      begin
        if Item.Image is TBitmap then
        begin
          TBitmap(Item.Image).Transparent := Transparent;
          TBitmap(Item.Image).TransparentMode := TransparentMode;
          TBitmap(Item.Image).TransparentColor := TransparentColor;
        end;
        InsertItem(T.Value, TCustomRVItemInfo(Item));
      end;
    finally
      Pic.Free;
    end;
  end;

  procedure ReadBreak(T: TXMLTag);
  var
    I: Integer;
    S: string;
    Item: TRVBreakItemInfo;
  begin
    Item := TRVBreakItemInfo.Create(RVData);
    Item.StyleNo := rvsBreak;
    Item.LineWidth := Default^.BreakWidth;
    Item.Color := Default^.BreakColor;
    for I := 0 to T.Attr.Count - 1 do
    begin
      S := LowerCase(T.Attr[I].Name);

      if ReadNonTextItemProperties(S, T.Attr[I], Item) then
        continue;

      if S = 'width' then Item.LineWidth := StrToIntDef(T.Attr[I].Value, Item.LineWidth)

      else if S = 'color' then Item.Color := StringToColor(T.Attr[I].Value);

    end;
    InsertItem(T.Value, TCustomRVItemInfo(Item));
  end;

  procedure ReadBullet(T: TXMLTag);
  var
    I: Integer;
    S: string;
    Item: TRVBulletItemInfo;
  begin
    Item := TRVBulletItemInfo.Create(RVData);
    Item.StyleNo := rvsBullet;
    Item.ParaNo := Default^.ParaNo;
    Item.ImageIndex := 0;
    Item.VAlign := Default^.VAlign;
    for I := 0 to T.Attr.Count - 1 do
    begin
      S := LowerCase(T.Attr[I].Name);

      if ReadRectItemProperties(S, T.Attr[I], Item) then
        continue;

      if S = 'img' then
        Item.ImageList := Bullets.Items[StrToInt(T.Attr[I].Value)];

    end;

    if Item.ParaNo = -1 then
      Item.ParaNo := Default^.ParaNo;

    Item.ParaNo := ParaMap[Item.ParaNo];

    InsertItem(T.Value, TCustomRVItemInfo(Item));
  end;

  procedure ReadHotSpot(T: TXMLTag);
  var
    ImgList: TImageList;
    Img, HotImg: TBitmap;
    Stream: TMemoryStream;
    S: string;
    I: Integer;
    Item: TRVHotSpotItemInfo;
  begin
    Item := TRVHotSpotItemInfo.Create(RVData);
    Item.StyleNo := rvsHotSpot;
    Item.ParaNo := Default^.ParaNo;
    Item.VAlign := Default^.VAlign;
    Stream := TMemoryStream.Create;
    Img := nil;
    HotImg := nil;
    for I := 0 to T.Attr.Count - 1 do
    begin
      S := LowerCase(T.Attr[I].Name);

      if ReadRectItemProperties(S, T.Attr[I], Item) then
        continue;

      if S = 'img' then
      begin
        Stream.Clear;
        RVFTextString2Stream(T.Attr[I].Value, Stream);
        Stream.Position := 0;
        Img := TBitmap.Create;
        Img.LoadFromStream(Stream);
      end

      else if S = 'hotimg' then
      begin
        Stream.Clear;
        RVFTextString2Stream(T.Attr[I].Value, Stream);
        Stream.Position := 0;
        HotImg := TBitmap.Create;
        HotImg.LoadFromStream(Stream);
      end

      else if S = 'imgsource' then
      begin
        Img := TBitmap.Create;
        Img.LoadFromFile(T.Root.BaseFilePath + T.Attr[I].Value);
      end

      else if S = 'hotimgsource' then
      begin
        HotImg := TBitmap.Create;
        HotImg.LoadFromFile(T.Root.BaseFilePath + T.Attr[I].Value);
      end

    end;

    if Item.ParaNo = -1 then
      Item.ParaNo := Default^.ParaNo;

    Item.ParaNo := ParaMap[Item.ParaNo];

    if Img <> nil then
    begin
      ImgList := TImageList.CreateSize(Img.Width, Img.Height);
      ImgList.AddMasked(Img, Img.Canvas.Pixels[0, 0]);
      Item.ImageList := ImgList;
      Item.ImageIndex := 0;
      Item.HotImageIndex := 0;
      if HotImg <> nil then
      begin
        ImgList.AddMasked(HotImg, HotImg.Canvas.Pixels[0, 0]);
        Item.HotImageIndex := 1;
      end;
      Img.Free;
    end;
    HotImg.Free;
    InsertItem(T.Value, TCustomRVItemInfo(Item));
    Stream.Free;
  end;

  procedure ReadControl(T: TXMLTag);
  var
    S: string;
    Stream, OStream: TStringStream;
    I: Integer;
    Item: TRVControlItemInfo;
    CtrlClass: TControlClass;

  begin
    Item := TRVControlItemInfo.Create(RVData);
    Item.StyleNo := rvsComponent;
    Item.ParaNo := Default^.ParaNo;
    Item.VAlign := Default^.VAlign;
    CtrlClass := nil;
    OStream := nil;
    for I := 0 to T.Attr.Count - 1 do
    begin
      S := LowerCase(T.Attr[I].Name);

      if ReadRectItemProperties(S, T.Attr[I], Item) then
        continue;

      if S = 'data' then
      begin
        if Assigned(OnLoadControl) then
        begin
          if Item.Control <> nil then
            Item.Control.Free;
          Item.Control := OnLoadControl(T.Attr[I].Value)
        end else
        begin
          Stream := TStringStream.Create(T.Attr[I].Value);
          OStream := TStringStream.Create('');
          try
            try
              ObjectTextToBinary(Stream, OStream);
              OStream.Position := 0;
            except
              OStream.Free;
              OStream := nil;
            end;
          finally
            Stream.Free;
          end;
        end;
      end

      else if S = 'class' then CtrlClass := TControlClass(FindClass(T.Attr[I].Value))

    end;

    if (OStream <> nil) and (Item.Control = nil) and (CtrlClass <> nil) then
    begin
      Item.Control := CtrlClass.Create(nil);
      Item.Control.Parent := RVData.GetParentControl;
      try
        OStream.ReadComponent(Item.Control);
      finally
        OStream.Free;
      end;
    end;

    if Item.Control <> nil then
    begin

      if Item.ParaNo = -1 then
        Item.ParaNo := Default^.ParaNo;

      Item.ParaNo := ParaMap[Item.ParaNo];

      InsertItem(T.Value, TCustomRVItemInfo(Item));
    end;
  end;

  procedure ReadTable(T: TXMLTag);
  var
    Item: TRVTableItemInfo;
    I, J, X, Y, RowSpan, ColSpan, Rows, Cols: Integer;
    S, S2: string;
    Cell: TRVTableCellData;
    A: TXMLAttr;
    gr: TGraphic;
    pic: TPicture;
    Stream: TMemoryStream;

    CellFirstItem, CellFullReformat: Boolean;
    CellInsertPosition: Integer;
    CellNonFirstItemsAdded: Integer;

    procedure SetTableOption(Attr: TXMLAttr; Option: TRVTableOption);
    begin
      if Attr.Value = '1' then
        Item.Options := Item.Options + [Option]
      else if Attr.Value = '0' then
        Item.Options := Item.Options - [Option];
    end;
    procedure SetTablePrintOption(Attr: TXMLAttr; Option: TRVTablePrintOption);
    begin
      if Attr.Value = '1' then
        Item.PrintOptions := Item.PrintOptions + [Option]
      else if Attr.Value = '0' then
        Item.PrintOptions := Item.PrintOptions - [Option];
    end;

  begin
    A := T.Attr.FindAttrOfName('rows');
    if A = nil then Exit;
    Rows := StrToIntDef(A.Value, 0);
    if Rows = 0 then Exit;

    A := T.Attr.FindAttrOfName('cols');
    if A = nil then Exit;
    Cols := StrToIntDef(A.Value, 0);
    if Cols = 0 then Exit;

    gr := nil;

    Item := TRVTableItemInfo.CreateEx(Rows, Cols, RVData);
    for I := 0 to T.Attr.Count - 1 do
    begin
      S := LowerCase(T.Attr[I].Name);

      if ReadNonTextItemProperties(S, T.Attr[I], Item) then
        continue;

      if S = 'width' then Item.BestWidth := StrToIntDef(T.Attr[I].Value, Item.BestWidth)

      else if S = 'bordercolor' then Item.BorderColor := StringToColor(T.Attr[I].Value)

      else if S = 'borderlightcolor' then Item.BorderLightColor := StringToColor(T.Attr[I].Value)

      else if S = 'borderstyle' then
      begin
        S2 := LowerCase(T.Attr[I].Value);
        if S2 = 'raised' then
          Item.BorderStyle := rvtbRaised
        else if S2 = 'lowered' then
          Item.BorderStyle := rvtbLowered
        else if S2 = 'color' then
          Item.BorderStyle := rvtbColor
        else if S2 = 'raisedcolor' then
          Item.BorderStyle := rvtbRaisedColor
        else if S2 = 'loweredcolor' then
          Item.BorderStyle := rvtbLoweredColor;
      end

      else if S = 'borderhspacing' then Item.BorderHSpacing := StrToIntDef(T.Attr[I].Value,
        Item.BorderHSpacing)      

      else if S = 'bordervspacing' then Item.BorderVSpacing := StrToIntDef(T.Attr[I].Value,
        Item.BorderVSpacing)

      else if S = 'borderwidth' then Item.BorderWidth := StrToIntDef(T.Attr[I].Value,
        Item.BorderWidth)

      else if S = 'cellbordercolor' then Item.CellBorderColor := StringToColor(T.Attr[I].Value)

      else if S = 'cellborderlightcolor' then Item.CellBorderLightColor := StringToColor(T.Attr[I].Value)

      else if S = 'cellborderstyle' then
      begin
        S2 := LowerCase(T.Attr[I].Value);
        if S2 = 'raised' then
          Item.CellBorderStyle := rvtbRaised
        else if S2 = 'lowered' then
          Item.CellBorderStyle := rvtbLowered
        else if S2 = 'color' then
          Item.CellBorderStyle := rvtbColor
        else if S2 = 'raisedcolor' then
          Item.CellBorderStyle := rvtbRaisedColor
        else if S2 = 'loweredcolor' then
          Item.CellBorderStyle := rvtbLoweredColor;
      end

      else if S = 'cellborderwidth' then Item.CellBorderWidth := StrToIntDef(T.Attr[I].Value,
        Item.CellBorderWidth)

      else if S = 'cellhspacing' then Item.CellHSpacing := StrToIntDef(T.Attr[I].Value,
        Item.CellHSpacing)

      else if S = 'cellpadding' then Item.CellPadding := StrToIntDef(T.Attr[I].Value,
        Item.CellPadding)

      else if S = 'cellvspacing' then Item.CellVSpacing := StrToIntDef(T.Attr[I].Value,
        Item.CellVSpacing)

      else if S = 'color' then Item.Color := StringToColor(T.Attr[I].Value)

      else if S = 'houtermostrule' then
      begin
        if T.Attr[I].Value = '1' then
          Item.HOutermostRule := True
        else if T.Attr[I].Value = '0' then
          Item.HOutermostRule := False;
      end

      else if S = 'hrulecolor' then
        Item.HRuleColor := StringToColor(T.Attr[I].Value)

      else if S = 'hrulewidth' then
        Item.HRuleWidth := StrToIntDef(T.Attr[I].Value, Item.HRuleWidth)

      else if S = 'editing' then
        SetTableOption(T.Attr[I], rvtoEditing)

      else if S = 'rowsizing' then
        SetTableOption(T.Attr[I], rvtoRowSizing)

      else if S = 'colsizing' then
        SetTableOption(T.Attr[I], rvtoColSizing)

      else if S = 'rowselect' then
        SetTableOption(T.Attr[I], rvtoRowSelect)

      else if S = 'colselect' then
        SetTableOption(T.Attr[I], rvtoColSelect)

      else if S = 'rtfsavecellwidth' then
        SetTableOption(T.Attr[I], rvtoRTFSaveCellPixelBestWidth)

      else if S = 'rtfallowautofit' then
        SetTableOption(T.Attr[I], rvtoRTFAllowAutofit)

      else if S = 'hidegridlines' then
        SetTableOption(T.Attr[I], rvtoHideGridLines)

      else if S = 'cellbelowborder' then
        SetTableOption(T.Attr[I], rvtoCellBelowBorders)

      else if S = 'overlappingcorners' then
        SetTableOption(T.Attr[I], rvtoOverlappingCorners)

      else if S = 'ignorecontentwidth' then
        SetTableOption(T.Attr[I], rvtoIgnoreContentWidth)

      else if S = 'ignorecontentheight' then
        SetTableOption(T.Attr[I], rvtoIgnoreContentHeight)

      else if S = 'printhalftoneborders' then
        SetTablePrintOption(T.Attr[I], rvtoHalftoneBorders)

      else if S = 'printsplitrows' then
        SetTablePrintOption(T.Attr[I], rvtoRowsSplit)

      else if S = 'printwhiteback' then
        SetTablePrintOption(T.Attr[I], rvtoWhiteBackground)

      else if S = 'textsavecolsep' then
        Item.TextColSeparator := T.Attr[I].Value

      else if S = 'textsaverowsep' then
        Item.TextRowSeparator := T.Attr[I].Value

      else if S = 'voutermostrule' then
      begin
        if T.Attr[I].Value = '1' then
          Item.VOutermostRule := True
        else if T.Attr[I].Value = '0' then
          Item.VOutermostRule := False;
      end

      else if S = 'vrulecolor' then
        Item.VRuleColor := StringToColor(T.Attr[I].Value)

      else if S = 'vrulewidth' then
        Item.VRuleWidth := StrToIntDef(T.Attr[I].Value, Item.VRuleWidth)

      else if S = 'headerrowcount' then
        Item.HeadingRowCount := StrToIntDef(T.Attr[I].Value, Item.HeadingRowCount)

      else if S = 'imagefilename' then
        Item.BackgroundImageFileName := T.Attr[I].Value

      else if s = 'backstyle' then
      begin
        if T.Attr[I].Value = 'color' then
          Item.BackgroundStyle := rvbsColor
        else if T.Attr[I].Value = 'stretch' then
          Item.BackgroundStyle := rvbsStretched
        else if T.Attr[I].Value = 'tile' then
          Item.BackgroundStyle := rvbsTiled
        else if T.Attr[I].Value = 'center' then
          Item.BackgroundStyle := rvbsCentered;
      end

      else if S = 'backimgclass' then
      begin
        gr := RV_CreateGraphics(TGraphicClass(GetClass(T.Attr[I].Value)));
        Item.BackgroundImage := gr;
        gr.Free;
        gr := nil;
      end

      else if S = 'backimg' then
      begin
        Stream := TMemoryStream.Create;
        try
          RVFTextString2Stream(T.Attr[I].Value, Stream);
          Stream.Position := 0;
          Item.BackgroundImage.LoadFromStream(Stream);
        finally
          Stream.Free;
        end;
      end

      else if S = 'backimgsource' then
      begin
        Pic := TPicture.Create;
        try
          Pic.LoadFromFile(T.Root.BaseFilePath + T.Attr[I].Value);
          Item.BackgroundImage := Pic.Graphic;
        finally
          Pic.Free;
        end;
      end

    end;

    for X := 0 to T.Items.Count - 1 do
    begin
      if LowerCase(T.Items[X].Name) = 'row' then
      begin
        for J := 0 to T.Items[X].Attr.Count - 1 do
        begin
          A := T.Items[X].Attr[J];
          if LowerCase(A.Name) = 'valign' then
          begin
            if LowerCase(A.Value) = 'top' then
              Item.Rows[X].VAlign := rvcTop
            else if LowerCase(A.Value) = 'middle' then
              Item.Rows[X].VAlign := rvcMiddle
            else if LowerCase(A.Value) = 'bottom' then
              Item.Rows[X].VAlign := rvcBottom
            else if LowerCase(A.Value) = 'default' then
              Item.Rows[X].VAlign := rvcVDefault;
          end;
        end;
        
        for Y := 0 to T.Items[X].Items.Count - 1 do
        begin
          if LowerCase(T.Items[X].Items[Y].Name) = 'cell' then
          begin
            if Item.Rows[X].Items[Y] <> nil then
            begin
              Cell := Item.Rows[X].Items[Y];
              RowSpan := 1;
              ColSpan := 1;
              for I := 0 to T.Items[X].Items[Y].Attr.Count - 1 do
              begin
                A := T.Items[X].Items[Y].Attr[I];
                S := LowerCase(A.Name);

                if S = 'height' then
                  Cell.BestHeight := StrToIntDef(A.Value, Cell.BestHeight)

                else if S = 'width' then
                  Cell.BestWidth := StrToIntDef(A.Value, Cell.BestWidth)

                else if S = 'color' then
                  Cell.Color := StringToColor(A.Value)

                else if S = 'bordercolor' then
                  Cell.BorderColor := StringToColor(A.Value)

                else if S = 'borderlightcolor' then
                  Cell.BorderLightColor := StringToColor(A.Value)

                else if S = 'colspan' then
                  ColSpan := StrToIntDef(A.Value, ColSpan)

                else if S = 'rowspan' then
                  RowSpan := StrToIntDef(A.Value, RowSpan)

                else if S = 'valign' then
                begin
                  if LowerCase(A.Value) = 'top' then
                    Cell.VAlign := rvcTop
                  else if LowerCase(A.Value) = 'middle' then
                    Cell.VAlign := rvcMiddle
                  else if LowerCase(A.Value) = 'bottom' then
                    Cell.VAlign := rvcBottom
                  else if LowerCase(A.Value) = 'default' then
                    Cell.VAlign := rvcVDefault;
                end

                else if S = 'imagefilename' then
                  Cell.BackgroundImageFileName := A.Value

                else if S = 'backimgclass' then
                begin
                  gr := RV_CreateGraphics(TGraphicClass(GetClass(A.Value)));
                end

                else if s = 'backstyle' then
                begin
                  if A.Value = 'color' then
                    Cell.BackgroundStyle := rvbsColor
                  else if A.Value = 'stretch' then
                    Cell.BackgroundStyle := rvbsStretched
                  else if A.Value = 'tile' then
                    Cell.BackgroundStyle := rvbsTiled
                  else if A.Value = 'center' then
                    Cell.BackgroundStyle := rvbsCentered;
                end

                else if S = 'backimg' then
                begin
                  if gr<>nil then begin
                    Stream := TMemoryStream.Create;
                    try
                      RVFTextString2Stream(A.Value, Stream);
                      Stream.Position := 0;
                      gr.LoadFromStream(Stream);
                      Cell.BackgroundImage := gr;
                      gr.Free;
                      gr := nil;
                    finally
                      Stream.Free;
                    end;
                  end;
                end

                else if S = 'backimgsource' then
                begin
                  Pic := TPicture.Create;
                  try
                    Pic.LoadFromFile(T.Root.BaseFilePath + A.Value);
                    Cell.BackgroundImage := Pic.Graphic;
                  finally
                    Pic.Free;
                  end;
                end

                else if S = 'leftbordervisible' then
                begin
                  if A.Value = '1' then
                    Cell.VisibleBorders.Left := True
                  else if A.Value = '0' then
                    Cell.VisibleBorders.Left := False;
                end

                else if S = 'rightbordervisible' then
                begin
                  if A.Value = '1' then
                    Cell.VisibleBorders.Right := True
                  else if A.Value = '0' then
                    Cell.VisibleBorders.Right := False;
                end

                else if S = 'topbordervisible' then
                begin
                  if A.Value = '1' then
                    Cell.VisibleBorders.Top := True
                  else if A.Value = '0' then
                    Cell.VisibleBorders.Top := False;
                end

                else if S = 'bottombordervisible' then
                begin
                  if A.Value = '1' then
                    Cell.VisibleBorders.Bottom := True
                  else if A.Value = '0' then
                    Cell.VisibleBorders.Bottom := False;
                end

              end;
              gr.Free;
              gr := nil;
              if (RowSpan > 1) or (ColSpan > 1) then
                Item.MergeCells(X, Y, ColSpan, RowSpan, True);
              Item.Cells[X, Y].GetRVData.Clear;
              CellFirstItem := False;
              CellFullReformat := False;
              CellInsertPosition := 0;
              Result := LoadRVDataFromXML(Cell.GetRVData, T.Items[X].Items[Y], Bullets, Default, CellInsertPosition, PaletteMode, RVPalette,
                RVLogPalette, PrevCP, StyleMap, ParaMap, ListMap, OnUnknownTag, OnloadControl, RVStyle, False,
                CellFirstItem, CellFullReformat,CellNonFirstItemsAdded, TagsArePChars);
//              if not Result then Exit;
            end;
          end;
        end;
      end;
    end;

    if Item.ParaNo = -1 then
      Item.ParaNo := Default^.ParaNo;

    Item.ParaNo := ParaMap[Item.ParaNo];

    InsertItem(T.Value, TCustomRVItemInfo(Item));
  end;

  procedure ReadMarker(T: TXMLTag);
  var
    item: TRVMarkerItemInfo;
    I: Integer;
    S: string;
  begin
    item := TRVMarkerItemInfo.CreateEx(RVData, -1, 0, 1, False);

    for I := 0 to T.Attr.Count - 1 do
      with T.Attr[I] do
      begin
        S := LowerCase(Name);

        if ReadRectItemProperties(S, T.Attr[I], Item) then
          continue;

        if S = 'htmlimagesize' then
        begin
          if T.Attr[I].Value = '1' then
            Item.NoHTMLImageSize := False
          else if T.Attr[I].Value = '0' then
            Item.NoHTMLImageSize := True;
        end

        else if S = 'style' then
          item.ListNo := FindNamedListNo(Value)

        else if S = 'styleno' then
          item.ListNo := StrToIntDef(Value, item.ListNo)

        else if S = 'level' then
          item.Level := StrToIntDef(Value, item.Level)

        else if S = 'start' then
          item.StartFrom := StrToIntDef(Value, item.StartFrom)

        else if S = 'usestart' then
        begin
          if Value = '1' then
            item.Reset := True
          else if Value = '0' then
            item.Reset := False;
        end;
      end;

      if Item.ParaNo = -1 then
        Item.ParaNo := Default^.ParaNo;

      Item.ParaNo := ParaMap[Item.ParaNo];
      if Item.ListNo>=0 then
        Item.ListNo := ListMap[Item.ListNo];

      InsertItem(T.Value, TCustomRVItemInfo(Item));
  end;

  procedure ReadTab(T: TXMLTag);
  var
    item: TRVTabItemInfo;
    I: Integer;
    S: string;
  begin
    item := TRVTabItemInfo.Create (RVData);
    item.StyleNo := rvsTab;

    for I := 0 to T.Attr.Count - 1 do
      with T.Attr[I] do
      begin
        S := LowerCase(Name);

        if ReadRectItemProperties(S, T.Attr[I], Item) then
          continue;

        if S = 'textstyle' then
          item.TextStyleNo := FindNamedStyleNo(Value)

        else if S =  'textstyleno' then
          item.TextStyleNo := StyleMap[StrToIntDef(Value, item.TextStyleNo)];
      end;

      if Item.ParaNo = -1 then
        Item.ParaNo := Default^.ParaNo;

      Item.ParaNo := ParaMap[Item.ParaNo];

      InsertItem(T.Value, TCustomRVItemInfo(Item));
  end;


  procedure ReadDefault(T: TXMLTag);
  var
    I: Integer;
    S: string;
  begin
    for I := 0 to T.Attr.Count - 1 do
    begin
      S := LowerCase(T.Attr[I].Name);

      if S = 'textstyleno' then
      begin
        Default^.StyleNo := StrToIntDef(T.Attr[I].Value, Default^.StyleNo);
      end

      else if S = 'parastyleno' then
      begin
        Default^.ParaNo := StrToIntDef(T.Attr[I].Value, Default^.ParaNo);
      end

      else if S = 'textstyle' then Default^.StyleNo := FindNamedStyleNo(T.Attr[I].Value)

      else if S = 'parastyle' then Default^.ParaNo := FindNamedParaNo(T.Attr[I].Value)

      else if S = 'valign' then
      begin
        if LowerCase(T.Attr[I].Value) = 'baseline' then
          Default^.VAlign := rvvaBaseLine
        else if LowerCase(T.Attr[I].Value) = 'middle' then
          Default^.VAlign := rvvaMiddle;
      end

      else if S = 'breakwidth' then Default^.BreakWidth := StrToIntDef(T.Attr[I].Value, Default^.BreakWidth)

      else if S = 'breakcolor' then Default^.BreakColor := StringToColor(T.Attr[I].Value);

    end;

    if Default^.StyleNo = -2 then
      Default^.StyleNo := 0;

    if Default^.ParaNo = -2 then
      Default^.ParaNo := 0;
  end;

var
  I: Integer;
  S: string;
  T: TXMLTag;
  PageBreak: Boolean;
  DefaultCreated: Boolean;
begin
  NonFirstItemsAdded := 0;
  Position := InsertPosition;
  Result := False;
  PrevCP := nil;
  DefaultCreated := Default = nil;
  if DefaultCreated then
  begin
    New(Default);
    Default^.StyleNo := 0;
    Default^.ParaNo := 0;
    Default^.VAlign := rvvaBaseLine;
    Default^.BreakColor := clNone;
    Default^.BreakWidth := 1;
    T := Node.Items.FindTagOfName('default');
    if T <> nil then
      ReadDefault(T);
  end;
  PageBreak := False;
  for I := 0 to Node.Items.Count - 1 do
  begin
    S := LowerCase(Node.Items[I].Name);
    if S = 'text' then ReadText(Node.Items[I])
    else if S = 'utext' then ReadUnicodeText(Node.Items[I])
    else if S = 'check' then ReadCheck(Node.Items[I])
    else if S = 'image' then ReadImage(Node.Items[I])
    else if S = 'hotimage' then ReadHotImage(Node.Items[I])
    else if S = 'break' then ReadBreak(Node.Items[I])
    else if S = 'bullet' then ReadBullet(Node.Items[I])
    else if S = 'hotspot' then ReadHotSpot(Node.Items[I])
    else if S = 'control' then ReadControl(Node.Items[I])
    else if S = 'table' then ReadTable(Node.Items[I])
    else if S = 'marker' then ReadMarker(Node.Items[I])
    else if S = 'tab' then ReadTab(Node.Items[I])
    else if S = 'default' then ReadDefault(Node.Items[I])
    else if S = 'pagebreak' then begin
      PageBreak := True;
      Continue;
    end else begin
      if Assigned(OnUnknownTag) then
        OnUnknownTag(Node.Items[I], RVData, Position);
      Inc(Position);
    end;
    if PageBreak then
      RVData.GetItem(Position-1).PageBreakBefore := True;
    PageBreak := False;
  end;
  if DefaultCreated then
    Dispose(Default);
end;

{------------------------------SaveRVDataToXML------------------------------}

procedure SaveRVDataToXML(RVData: TCustomRVData; Node: TXMLTag;
  Default: TRichViewXMLStyle; SaveStyle: Boolean; RVStyle: TRVStyle;
  SaveImgToFiles: Boolean; Bullets: TImgListList;
  SelStart, SelOffset, SelEnd, SelEndOffset: Integer;
  TagsArePChars: Boolean; OnUnknownItem: TRVXMLUnknownItemEvent;
  OnSaveControl: TRVXMLSaveControlEvent;
  UseStyleNames: Boolean);

  function GetTagStr(Tag: Integer): string;
  begin
    if Tag = 0 then
      Result := ''
    else
      Result := PChar(Tag);
  end;

  procedure WriteCheck(T: TXMLTag; ItemNo: Integer);
  var
    T2: TXMLTag;
    Tag, Name: string;
    Cp: TCheckpointData;
    IntTag: Integer;
    Vis: Boolean;
  begin
    Cp := RVData.GetItemCheckpoint(ItemNo);
    if Cp <> nil then
    begin
      T2 := T.Items.AddTag('check', T, T.Root);
      RVData.GetCheckpointInfo(Cp, IntTag, Name, Vis);

      if TagsArePChars then
      begin
        Tag := GetTagStr(IntTag);
        if Tag <> '' then
          T2.Attr.AddAttr('tag', Tag);
      end else
      begin
        if IntTag <> 0 then
          T2.Attr.AddAttr('tagnumber', IntToStr(IntTag));
      end;

      if Vis = False then
        T2.Attr.AddAttr('vis', '0');

      T2.Value := Name;
    end;
  end;

  procedure WriteItem(T: TXMLTag; Item: TCustomRVItemInfo; ItemNo: Integer);
  var StrValue: String;
  begin
    WriteCheck(T.Parent, ItemNo);
    if TagsArePChars then
    begin
      if GetTagStr(Item.Tag) <> '' then
        T.Attr.AddAttr('tag', GetTagStr(Item.Tag));
    end else
    begin
      if Item.Tag <> 0 then
        T.Attr.AddAttr('tagnumber', IntToStr(Item.Tag));
    end;
    {$IFNDEF RVDONOTUSEITEMHINTS}
    if Item.GetExtraStrProperty(rvespHint, StrValue) and (StrValue<>'') then
      T.Attr.AddAttr('hint', StrValue);
    {$ENDIF}
    if Item.SameAsPrev then
    begin
      T.Attr.AddAttr('br', '0')
    end else if Item.ParaNo <> Default.ParaNo then
    begin
      if UseStyleNames then
        T.Attr.AddAttr('parastyle', RVStyle.ParaStyles[Item.ParaNo].StyleName)
      else
        T.Attr.AddAttr('parastyleno', IntToStr(Item.ParaNo));
    end;
    if Item.BR and (not Item.SameAsPrev) then
      T.Attr.AddAttr('newpara', '0');
  end;

  procedure WriteText(T: TXMLTag; Item: TRVTextItemInfo; ItemNo: Integer);
  begin
    if UseStyleNames then
    begin
      if Item.StyleNo <> Default.StyleNo then
        T.Attr.AddAttr('textstyle', RVStyle.TextStyles[Item.StyleNo].StyleName);
    end else begin
      if Item.StyleNo <> Default.StyleNo then
        T.Attr.AddAttr('textstyleno', IntToStr(Item.StyleNo));
    end;
    WriteItem(T, Item, ItemNo);
    T.Value := RVData.Items[ItemNo];
  end;

  procedure WriteText2(T: TXMLTag; Item: TRVTextItemInfo; Text: string);
  begin
    if Item.StyleNo <> Default.StyleNo then
      T.Attr.AddAttr('textstyleno', IntToStr(Item.StyleNo));

    if TagsArePChars then
    begin
      if GetTagStr(Item.Tag) <> '' then
        T.Attr.AddAttr('tag', GetTagStr(Item.Tag));
    end else
    begin
      if Item.Tag <> 0 then
        T.Attr.AddAttr('tagnumber', IntToStr(Item.Tag));
    end;

    {$IFNDEF RVDONOTUSEITEMHINTS}
    if Item.Hint<>'' then
      T.Attr.AddAttr('hint', Item.Hint);
    {$ENDIF}

    if Item.SameAsPrev then
    begin
      T.Attr.AddAttr('br', '0')
    end else if Item.ParaNo <> Default.ParaNo then
    begin
      if UseStyleNames then
        T.Attr.AddAttr('parastyle', RVStyle.ParaStyles[Item.ParaNo].StyleName)
      else
        T.Attr.AddAttr('parastyleno', IntToStr(Item.ParaNo));
    end;

    if Item.BR and (not Item.SameAsPrev) then
      T.Attr.AddAttr('newpara', '0');

    T.Value := Text;
  end;

  procedure WriteNonTextItem(T: TXMLTag; Item: TCustomRVItemInfo; ItemNo: Integer);
  var IntValue: Integer;
  begin
    WriteItem(T, Item, ItemNo);
    if Item.GetExtraIntProperty(rvepDeleteProtect, IntValue) and (IntValue<>0) then
      T.Attr.AddAttr('deleteprotect', '1');
  end;

  procedure WriteRectItem(T: TXMLTag; Item: TRVRectItemInfo; ItemNo: Integer;
    SaveSpacing: Boolean);
  begin
    WriteNonTextItem(T, Item, ItemNo);

    if Item.VAlign <> Default.VAlign then
      if Item.VAlign = rvvaBaseLine then
        T.Attr.AddAttr('valign', 'baseline')
      else
        T.Attr.AddAttr('valign', 'middle');

    if Item.VShift <> 0 then
      T.Attr.AddAttr('vshift', IntToStr(Item.VShift));

    if Item.VShiftAbs then
      T.Attr.AddAttr('vshiftabs', '1');

    if SaveSpacing then
      T.Attr.AddAttr('spacing', IntToStr(Item.Spacing));
  end;


  procedure WriteBreak(T: TXMLTag; Item: TRVBreakItemInfo; ItemNo: Integer);
  begin
    WriteNonTextItem(T, Item, ItemNo);

    if Item.LineWidth <> Default.BreakWidth then
      T.Attr.AddAttr('width', IntToStr(Item.LineWidth));

    if Item.Color <> Default.BreakColor then
      T.Attr.AddAttr('color', ColorToString(Item.Color));
  end;

  procedure WriteImage(T: TXMLTag; Item: TRVGraphicItemInfo; ItemNo: Integer);
  var
    FileName: string;
    Stream: TMemoryStream;
  begin
    WriteRectItem(T, Item, ItemNo, True);

    if Item.ImageWidth <> 0 then
      T.Attr.AddAttr('width', IntToStr(Item.ImageWidth));

    if Item.ImageHeight <> 0 then
      T.Attr.AddAttr('height', IntToStr(Item.ImageHeight));

    if Item.MinHeightOnPage <> 0 then
      T.Attr.AddAttr('minheighonpage', IntToStr(Item.MinHeightOnPage));

    if Item.NoHTMLImageSize then
      T.Attr.AddAttr('htmlimagesize', '0');

    if Item.Alt<>'' then
      T.Attr.AddAttr('alt', Item.Alt);

    if Item.ImageFileName<>'' then
      T.Attr.AddAttr('imagefilename', Item.ImageFileName);

    if Item.Image is TBitmap then
      with TBitmap(Item.Image) do
        if Transparent then
        begin
          T.Attr.AddAttr('transparent', '1');
          if TransparentMode = tmFixed then
          begin
            T.Attr.AddAttr('transparentmode', 'fixed');
            if TransparentColor <> 0 then
              T.Attr.AddAttr('transparentcolor', ColorToString(TransparentColor));
          end;
        end;

    if SaveImgToFiles then
    begin

      if RVData.Items[ItemNo] = '' then
        FileName := GetImageFileName(T.Root.BaseFilePath, 'Image', GraphicExtension(TGraphicClass(Item.Image.ClassType)))
      else
        FileName := GetImageFileName(T.Root.BaseFilePath, RVData.Items[ItemNo],
          GraphicExtension(TGraphicClass(Item.Image.ClassType)));

      Item.Image.SaveToFile(FileName);
      T.Attr.AddAttr('imgsource', ExtractFileName(FileName));
    end else begin
      T.Attr.AddAttr('class', Item.Image.ClassName);

      Stream := TMemoryStream.Create;
      try
        Item.Image.SaveToStream(Stream);
        Stream.Position := 0;
        T.Attr.AddAttr('img', RVFStream2TextString(Stream));
      finally
        Stream.Free;
      end;
    end;

    T.Value := RVData.Items[ItemNo];
  end;

  procedure WriteBullet(T: TXMLTag; Item: TRVBulletItemInfo; ItemNo: Integer);
  begin
    WriteRectItem(T, Item, ItemNo, True);

    if Item.NoHTMLImageSize then
      T.Attr.AddAttr('htmlimagesize', '0');    

    T.Attr.AddAttr('img', IntToStr(Bullets.IndexOf(Item.ImageList)));

    T.Value := RVData.Items[ItemNo];
  end;


  procedure WriteHotSpot(T: TXMLTag; Item: TRVHotSpotItemInfo; ItemNo: Integer);
  var
    Stream: TMemoryStream;
    Bmp: TBitmap;
    FileName: string;
  begin
    WriteRectItem(T, Item, ItemNo, True);

    if Item.NoHTMLImageSize then
      T.Attr.AddAttr('htmlimagesize', '0');    

    Bmp := TBitmap.Create;
    Item.ImageList.GetBitmap(Item.ImageIndex, Bmp);
    if SaveImgToFiles then
    begin

      if RVData.Items[ItemNo] = '' then
        FileName := GetImageFileName(T.Root.BaseFilePath, 'HotSpot', 'bmp')
      else
        FileName := GetImageFileName(T.Root.BaseFilePath, RVData.Items[ItemNo], 'bmp');

      Bmp.SaveToFile(FileName);
      T.Attr.AddAttr('imgsource', ExtractFileName(FileName));

    end else begin

      Stream := TMemoryStream.Create;
      try
        Bmp.SaveToStream(Stream);
        Stream.Position := 0;
        T.Attr.AddAttr('img', RVFStream2TextString(Stream));
      finally
        Stream.Free;
      end;

    end;

    if Item.ImageIndex <> Item.HotImageIndex then
    begin
      Bmp := TBitmap.Create;
      Item.ImageList.GetBitmap(Item.HotImageIndex, Bmp);
      if SaveImgToFiles then
      begin

        if RVData.Items[ItemNo] = '' then
          FileName := GetImageFileName(T.Root.BaseFilePath, 'HotSpot', 'bmp')
        else
          FileName := GetImageFileName(T.Root.BaseFilePath, RVData.Items[ItemNo], 'bmp');

        Bmp.SaveToFile(FileName);
        T.Attr.AddAttr('hotimgsource', ExtractFileName(FileName));

      end else begin

        Stream := TMemoryStream.Create;
        try
          Bmp.SaveToStream(Stream);
          Stream.Position := 0;
          T.Attr.AddAttr('hotimg', RVFStream2TextString(Stream));
        finally
          Stream.Free;
        end;

      end;
    end;

    T.Value := RVData.Items[ItemNo];
  end;

  procedure WriteControl(T: TXMLTag; Item: TRVControlItemInfo; ItemNo: Integer);
  var
    Stream: TMemoryStream;
    OStream: TStringStream;
  begin
    WriteRectItem(T, Item, ItemNo, True);

    if Assigned(OnSaveControl) then
      T.Attr.AddAttr('data', OnSaveControl(Item.Control))
    else begin
      T.Attr.AddAttr('class', Item.Control.ClassName);
      Stream := TMemoryStream.Create;
      try
        Stream.WriteComponent(Item.Control);
        Stream.Position := 0;
        OStream := TStringStream.Create('');
        try
          ObjectBinaryToText(Stream, OStream);
          OStream.Position := 0;
//          T.Attr.AddAttr('data', StringReplace(OStream.DataString, crlf, ' ', [rfReplaceAll]));
          T.Attr.AddAttr('data', OStream.DataString);
        finally
          OStream.Free;
        end;
      finally
        Stream.Free;
      end;
    end;

    T.Value := RVData.Items[ItemNo];
  end;

  procedure WriteTable(T: TXMLTag; Item: TRVTableItemInfo; ItemNo: Integer);
  var
    X, Y: Integer;
    FileName: String;
    Stream: TMemoryStream;
    Cell: TRVTableCellData;
    CellTag: TXMLTag;
  begin
    WriteNonTextItem(T, Item, ItemNo);

    T.Attr.AddAttr('rows', IntToStr(Item.Rows.Count));
    T.Attr.AddAttr('cols', IntToStr(Item.Rows[0].Count));

    if Item.BestWidth <> 0 then
      T.Attr.AddAttr('width', IntToStr(Item.BestWidth));

    if Item.BorderColor <> clWindowText then
      T.Attr.AddAttr('bordercolor', ColorToString(Item.BorderColor));

    if Item.BorderLightColor <> clBtnHighlight then
      T.Attr.AddAttr('borderlightcolor', ColorToString(Item.BorderLightColor));

    case Item.BorderStyle of
      rvtbLowered:
        T.Attr.AddAttr('borderstyle', 'lowered');
      rvtbColor:
        T.Attr.AddAttr('borderstyle', 'color');
      rvtbRaisedColor:
        T.Attr.AddAttr('borderstyle', 'raisedcolor');
      rvtbLoweredColor:
        T.Attr.AddAttr('borderstyle', 'loweredcolor');
    end;

    if Item.BorderHSpacing <> 2 then
      T.Attr.AddAttr('borderhspacing', IntToStr(Item.BorderHSpacing));    

    if Item.BorderVSpacing <> 2 then
      T.Attr.AddAttr('bordervspacing', IntToStr(Item.BorderVSpacing));

    if Item.BorderWidth <> 0 then
      T.Attr.AddAttr('borderwidth', IntToStr(Item.BorderWidth));

    if Item.CellBorderColor <> clWindowText then
      T.Attr.AddAttr('cellbordercolor', ColorToString(Item.CellBorderColor));

    if Item.CellBorderLightColor <> clBtnHighlight then
      T.Attr.AddAttr('cellborderlightcolor', ColorToString(Item.CellBorderLightColor));

    case Item.CellBorderStyle of
      rvtbRaised:
        T.Attr.AddAttr('cellborderstyle', 'raised');
      rvtbColor:
          T.Attr.AddAttr('cellborderstyle', 'color');
      rvtbRaisedColor:
        T.Attr.AddAttr('cellborderstyle', 'raisedcolor');
      rvtbLoweredColor:
        T.Attr.AddAttr('cellborderstyle', 'loweredcolor');
    end;

    if Item.CellBorderWidth <> 0 then
      T.Attr.AddAttr('cellborderwidth', IntToStr(Item.CellBorderWidth));

    if Item.CellHSpacing <> 2 then
      T.Attr.AddAttr('cellhspacing', IntToStr(Item.CellHSpacing));

    if Item.CellPadding <> 1 then
      T.Attr.AddAttr('cellpadding', IntToStr(Item.CellPadding));

    if Item.CellVSpacing <> 2 then
      T.Attr.AddAttr('cellvspacing', IntToStr(Item.CellVSpacing));

    if Item.Color <> clWindow then
      T.Attr.AddAttr('color', ColorToString(Item.Color));

    if Item.HOutermostRule then
      T.Attr.AddAttr('houtermostrule', '1');

    if Item.HRuleColor <> clWindowText then
      T.Attr.AddAttr('hrulecolor', ColorToString(Item.HRuleColor));

    if Item.HRuleWidth <> 0 then
      T.Attr.AddAttr('hrulewidth', IntToStr(Item.HRuleWidth));

    if not (rvtoEditing in Item.Options) then
      T.Attr.AddAttr('editing', '0');

    if not (rvtoRowSizing in Item.Options) then
      T.Attr.AddAttr('rowsizing', '0');

    if not (rvtoColSizing in Item.Options) then
      T.Attr.AddAttr('colsizing', '0');

    if not (rvtoRowSelect in Item.Options) then
      T.Attr.AddAttr('rowselect', '0');

    if not (rvtoColSelect in Item.Options) then
      T.Attr.AddAttr('colselect', '0');

    if rvtoRTFSaveCellPixelBestWidth in Item.Options then
      T.Attr.AddAttr('rtfsavecellwidth', '1');

    if rvtoRTFAllowAutoFit in Item.Options then
      T.Attr.AddAttr('rtfallowautofit', '1');

    if rvtoHideGridLines in Item.Options then
      T.Attr.AddAttr('hidegridlines', '1');

    if rvtoCellBelowBorders in Item.Options then
      T.Attr.AddAttr('cellbelowborder', '1');

    if rvtoOverlappingCorners in Item.Options then
      T.Attr.AddAttr('overlappingcorners', '1');

    if rvtoIgnoreContentWidth in Item.Options then
      T.Attr.AddAttr('ignorecontentwidth', '1');

    if rvtoIgnoreContentHeight in Item.Options then
      T.Attr.AddAttr('ignorecontentheight', '1');

    if not (rvtoHalftoneBorders in Item.PrintOptions) then
      T.Attr.AddAttr('printhalftoneborders', '0');

    if not (rvtoRowsSplit in Item.PrintOptions) then
      T.Attr.AddAttr('printsplitrows', '0');

    if rvtoWhiteBackground in Item.PrintOptions then
      T.Attr.AddAttr('printwhiteback', '1');

    if Item.TextColSeparator <> #13#10 then
      T.Attr.AddAttr('textsavecolsep',Item.TextColSeparator);

    if Item.TextRowSeparator <> #13#10 then
      T.Attr.AddAttr('textsaverowsep',Item.TextRowSeparator);

    if Item.VOutermostRule then
      T.Attr.AddAttr('voutermostrule', '1');

    if Item.VRuleColor <> clWindowText then
      T.Attr.AddAttr('vrulecolor', ColorToString(Item.VRuleColor));

    if Item.VRuleWidth <> 0 then
      T.Attr.AddAttr('vrulewidth', IntToStr(Item.VRuleWidth));

    if Item.HeadingRowCount <> 0 then
      T.Attr.AddAttr('headerrowcount', IntToStr(Item.HeadingRowCount));

    if Item.BackgroundImageFileName<>'' then
      T.Attr.AddAttr('imagefilename', Item.BackgroundImageFileName);

    case Item.BackgroundStyle of
      rvbsStretched:
        T.Attr.AddAttr('backstyle', 'stretch');
      rvbsTiled:
        T.Attr.AddAttr('backstyle', 'tile');
      rvbsCentered:
        T.Attr.AddAttr('backstyle', 'center');
    end;

    if (Item.BackgroundImage<>nil) and not Item.BackgroundImage.Empty then begin
      if SaveImgToFiles then
      begin

        if RVData.Items[ItemNo] = '' then
          FileName := GetImageFileName(T.Root.BaseFilePath, 'bkgnd',
            GraphicExtension(TGraphicClass(Item.BackgroundImage.ClassType)))
        else
          FileName := GetImageFileName(T.Root.BaseFilePath, RVData.Items[ItemNo],
            GraphicExtension(TGraphicClass(Item.BackgroundImage.ClassType)));

        Item.BackgroundImage.SaveToFile(FileName);
        T.Attr.AddAttr('backimgsource', ExtractFileName(FileName));
      end else begin
        T.Attr.AddAttr('backimgclass', Item.BackgroundImage.ClassName);

        Stream := TMemoryStream.Create;
        try
          Item.BackgroundImage.SaveToStream(Stream);
          Stream.Position := 0;
          T.Attr.AddAttr('backimg', RVFStream2TextString(Stream));
        finally
          Stream.Free;
        end;
      end;
    end;

    for X := 0 to Item.Rows.Count - 1 do
    begin
      T.Items.AddTag('row', T, T.Root);
      if Item.Rows[X].VAlign <> rvcTop then
        if Item.Rows[X].VAlign = rvcMiddle then
          T.Items[X].Attr.AddAttr('valign', 'middle')
        else if Item.Rows[X].VAlign = rvcBottom then
          T.Items[X].Attr.AddAttr('valign', 'bottom')
        else if Item.Rows[X].VAlign = rvcVDefault then
          T.Items[X].Attr.AddAttr('valign', 'default');

      for Y := 0 to Item.Rows[0].Count - 1 do
      begin
        T.Items[X].Items.AddTag('cell', T.Items[X], T.Items[X].Root);
        Cell := Item.Cells[X, Y];
        if Cell <> nil then
        begin
          CellTag := T.Items[X].Items[Y];
          if Cell.RowSpan > 1 then
            CellTag.Attr.AddAttr('rowspan', IntToStr(Cell.RowSpan));

          if Cell.ColSpan > 1 then
            CellTag.Attr.AddAttr('colspan', IntToStr(Cell.ColSpan));

          if Cell.BestHeight <> 0 then
            CellTag.Attr.AddAttr('height', IntToStr(Cell.BestHeight));

          if Cell.BestWidth <> 0 then
            CellTag.Attr.AddAttr('width', IntToStr(Cell.BestWidth));

          if Cell.Color <> clNone then
            CellTag.Attr.AddAttr('color', ColorToString(Cell.Color));

          if Cell.BorderColor <> clNone then
            CellTag.Attr.AddAttr('bordercolor', ColorToString(Cell.BorderColor));

          if Cell.BorderLightColor <> clNone then
            CellTag.Attr.AddAttr('borderlightcolor', ColorToString(Cell.BorderLightColor));

          case Cell.VAlign of
            rvcMiddle:
              CellTag.Attr.AddAttr('valign', 'middle');
            rvcBottom:
              CellTag.Attr.AddAttr('valign', 'bottom');
            rvcTop:
              CellTag.Attr.AddAttr('valign', 'top');
          end;

          case Cell.BackgroundStyle of
            rvbsStretched:
              CellTag.Attr.AddAttr('backstyle', 'stretch');
            rvbsTiled:
              CellTag.Attr.AddAttr('backstyle', 'tile');
            rvbsCentered:
              CellTag.Attr.AddAttr('backstyle', 'center');
          end;

          if Cell.BackgroundImageFileName<>'' then
            CellTag.Attr.AddAttr('imagefilename', Cell.BackgroundImageFileName);

          if (Cell.BackgroundImage<>nil) and not Cell.BackgroundImage.Empty then begin
            if SaveImgToFiles then
            begin
              FileName := GetImageFileName(T.Root.BaseFilePath, 'bkgnd',
                GraphicExtension(TGraphicClass(Cell.BackgroundImage.ClassType)));
              Cell.BackgroundImage.SaveToFile(FileName);
              CellTag.Attr.AddAttr('backimgsource', ExtractFileName(FileName));
            end else begin
              CellTag.Attr.AddAttr('backimgclass', Cell.BackgroundImage.ClassName);

              Stream := TMemoryStream.Create;
              try
                Cell.BackgroundImage.SaveToStream(Stream);
                Stream.Position := 0;
                CellTag.Attr.AddAttr('backimg', RVFStream2TextString(Stream));
              finally
                Stream.Free;
              end;
            end;
          end;

          if not Cell.VisibleBorders.Left then
            CellTag.Attr.AddAttr('leftbordervisible', '0');

          if not Cell.VisibleBorders.Right then
            CellTag.Attr.AddAttr('rightbordervisible', '0');

          if not Cell.VisibleBorders.Top then
            CellTag.Attr.AddAttr('topbordervisible', '0');

          if not Cell.VisibleBorders.Bottom then
            CellTag.Attr.AddAttr('bottombordervisible', '0');

          SaveRVDataToXML(Cell.GetRVData, CellTag, Default, False, RVStyle,
            SaveImgToFiles, Bullets, 0, 0, Item.Cells[X, Y].GetRVData.Items.Count, 0,
            TagsArePChars, OnUnknownItem, OnSaveControl, UseStyleNames);
        end;
      end;
    end;
  end;

  procedure WriteMarker(T: TXMLTag; Item: TRVMarkerItemInfo; ItemNo: Integer);
  begin
    WriteRectItem(T, Item, ItemNo, False);
    with T.Attr do
    begin
      if UseStyleNames then
        AddAttr('style', RVStyle.ListStyles[Item.ListNo].StyleName)
      else
        AddAttr('styleno', IntToStr(Item.ListNo));

      if Item.Level <> 0 then
        AddAttr('level', IntToStr(Item.Level));

      if Item.NoHTMLImageSize then
        AddAttr('htmlimagesize', '0');

      if Item.StartFrom <> 0 then
        AddAttr('start', IntToStr(Item.StartFrom));

      if Item.Reset = True then
        AddAttr('usestart', '1');
    end;
  end;

  procedure WriteTab(T: TXMLTag; Item: TRVTabItemInfo; ItemNo: Integer);
  begin
    WriteRectItem(T, Item, ItemNo, False);
    with T.Attr do
    begin
      if UseStyleNames then
        AddAttr('textstyle', RVStyle.TextStyles[Item.TextStyleNo].StyleName)
      else
        AddAttr('textstyleno', IntToStr(Item.TextStyleNo));
    end;
  end;

var
  I, Item: Integer;
  T: TXMLTag;
  TextInfo: TRVTextItemInfo;
begin
  if SaveStyle then
    if (Default.BreakColor <> DefaultXMLStyle.BreakColor) or
      (Default.BreakWidth <> DefaultXMLStyle.BreakWidth) or
      (Default.ParaNo <> DefaultXMLStyle.ParaNo) or
      (Default.StyleNo <> DefaultXMLStyle.StyleNo) or
      (Default.VAlign <> DefaultXMLStyle.VAlign) then
    begin
      T := Node.Items.AddTag('default', Node, Node.Root);
      if Default.StyleNo <> 0 then
        T.Attr.AddAttr('textstyleno', IntToStr(Default.StyleNo));
      if Default.ParaNo <> 0 then
        T.Attr.AddAttr('parastyleno', IntToStr(Default.ParaNo));
      if Default.VAlign <> rvvaBaseLine then
        T.Attr.AddAttr('valign', 'middle');
      if Default.BreakColor <> clNone then
        T.Attr.AddAttr('breakcolor', ColorToString(Default.BreakColor));
      if Default.BreakWidth <> 1 then
        T.Attr.AddAttr('breakwidth', IntToStr(Default.BreakWidth));
      end;

  if (SelStart = SelEnd) and (SelStart <> 0) and (SelEnd <> 0) and (SelOffset <> SelEndOffset) then
  begin
    if RVData.GetItem(SelStart) is TRVTextItemInfo then
    begin
      TextInfo := TRVTextItemInfo.Create(RVData);
      TextInfo.Assign(TRVTextItemInfo(RVData.GetItem(SelStart)));
      WriteText2(Node.Items.AddTag('text', Node, Node.Root), TextInfo,
        Copy(RVData.Items[SelStart], SelOffset + 1, SelEndOffset - SelOffset));
    end else
    begin
      case RVData.GetItemStyle(SelEnd) of
        rvsBreak: begin
          T := Node.Items.AddTag('break', Node, Node.Root);
          WriteBreak(T, TRVBreakItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
        rvsPicture: begin
          T := Node.Items.AddTag('image', Node, Node.Root);
          WriteImage(T, TRVGraphicItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
        rvsHotspot: begin
          T := Node.Items.AddTag('hotspot', Node, Node.Root);
          WriteHotSpot(T, TRVHotSpotItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
        rvsComponent: begin
          T := Node.Items.AddTag('control', Node, Node.Root);
          WriteControl(T, TRVControlItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
        rvsBullet: begin
          T := Node.Items.AddTag('bullet', Node, Node.Root);
          WriteBullet(T, TRVBulletItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
        rvsHotPicture: begin
          T := Node.Items.AddTag('hotimage', Node, Node.Root);
          WriteImage(T, TRVGraphicItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
        rvsTable: begin
          T := Node.Items.AddTag('table', Node, Node.Root);
          WriteTable(T, TRVTableItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
        rvsListMarker: begin
          T := Node.Items.AddTag('marker', Node, Node.Root);
          WriteMarker(T, TRVMarkerItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
        rvsTab: begin
          T := Node.Items.AddTag('tab', Node, Node.Root);
          WriteTab(T, TRVTabItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
        else begin
          if Assigned(OnUnknownItem) then
            OnUnknownItem(RVData, SelEnd, TCustomRVItemInfo(RVData.GetItem(SelEnd)),
              Node.Items.AddTag('', Node, Node.Root));
        end;
      end;
    end;
    Exit;
  end;

  if SelOffset <> 0 then
  begin
    if RVData.GetItem(SelStart) is TRVTextItemInfo then
    begin
      TextInfo := TRVTextItemInfo.Create(RVData);
      TextInfo.Assign(TRVTextItemInfo(RVData.GetItem(SelStart)));
      WriteText2(Node.Items.AddTag('text', Node, Node.Root), TextInfo,
        Copy(RVData.Items[SelStart], SelOffset + 1, Length(RVData.Items[SelStart]) - SelOffset));
    end;
  end;

  for I := SelStart to SelEnd - 1 do
  begin
    Item := RVData.GetItemStyle(I);
    if RVData.GetItem(i).PageBreakBefore then
      Node.Items.AddTag('pagebreak', Node, Node.Root);
    if Item >= 0 then
    begin
      if RVStyle.TextStyles[Item].Unicode then
      begin
        T := Node.Items.AddTag('utext', Node, Node.Root);
        T.Unicode := True;
        WriteText(T, TRVTextItemInfo(RVData.GetItem(i)), I);
      end else
      begin
        T := Node.Items.AddTag('text', Node, Node.Root);
        WriteText(T, TRVTextItemInfo(RVData.GetItem(i)), I);
      end;
    end else
    case Item of
      rvsBreak: begin
        T := Node.Items.AddTag('break', Node, Node.Root);
        WriteBreak(T, TRVBreakItemInfo(RVData.GetItem(i)), I); end;
      rvsPicture: begin
        T := Node.Items.AddTag('image', Node, Node.Root);
        WriteImage(T, TRVGraphicItemInfo(RVData.GetItem(i)), I); end;
      rvsHotspot: begin
        T := Node.Items.AddTag('hotspot', Node, Node.Root);
        WriteHotSpot(T, TRVHotSpotItemInfo(RVData.GetItem(i)), I); end;
      rvsComponent: begin
        T := Node.Items.AddTag('control', Node, Node.Root);
        WriteControl(T, TRVControlItemInfo(RVData.GetItem(i)), I); end;
      rvsBullet: begin
        T := Node.Items.AddTag('bullet', Node, Node.Root);
        WriteBullet(T, TRVBulletItemInfo(RVData.GetItem(i)), I); end;
      rvsHotPicture: begin
        T := Node.Items.AddTag('hotimage', Node, Node.Root);
        WriteImage(T, TRVGraphicItemInfo(RVData.GetItem(i)), I); end;
      rvsTable: begin
        T := Node.Items.AddTag('table', Node, Node.Root);
        WriteTable(T, TRVTableItemInfo(RVData.GetItem(i)), I); end;
      rvsListMarker: begin
        T := Node.Items.AddTag('marker', Node, Node.Root);
        WriteMarker(T, TRVMarkerItemInfo(RVData.GetItem(i)), I); end;
      rvsTab: begin
        T := Node.Items.AddTag('tab', Node, Node.Root);
        WriteTab(T, TRVTabItemInfo(RVData.GetItem(i)), I); end;
      else begin
        if Assigned(OnUnknownItem) then
          OnUnknownItem(RVData, I, RVData.GetItem(i),
            Node.Items.AddTag('', Node, Node.Root));
      end;
    end;
  end;

  if SelEnd >= RVData.Items.Count then Exit;

  if RVData.GetItem(SelEnd) is TRVTextItemInfo then
  begin
    if SelEndOffset <> 0 then
    begin
      TextInfo := TRVTextItemInfo.Create(RVData);
      TextInfo.Assign(TRVTextItemInfo(RVData.GetItem(SelEnd)));
      if not RVStyle.TextStyles[TextInfo.StyleNo].Unicode then
        WriteText2(Node.Items.AddTag('text', Node, Node.Root), TextInfo,
          Copy(RVData.Items[SelEnd], 1, SelEndOffset))
      else begin
        T := Node.Items.AddTag('utext', Node, Node.Root);
        T.Unicode := True;
        WriteText2(T, TextInfo, Copy(RVData.Items[SelEnd], 1, SelEndOffset * 2));
      end;
    end;
  end else
  begin
    case RVData.GetItemStyle(SelEnd) of
      rvsBreak: begin
        T := Node.Items.AddTag('break', Node, Node.Root);
        WriteBreak(T, TRVBreakItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
      rvsPicture: begin
        T := Node.Items.AddTag('image', Node, Node.Root);
        WriteImage(T, TRVGraphicItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
      rvsHotspot: begin
        T := Node.Items.AddTag('hotspot', Node, Node.Root);
        WriteHotSpot(T, TRVHotSpotItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
      rvsComponent: begin
        T := Node.Items.AddTag('control', Node, Node.Root);
        WriteControl(T, TRVControlItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
      rvsBullet: begin
        T := Node.Items.AddTag('bullet', Node, Node.Root);
        WriteBullet(T, TRVBulletItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
      rvsHotPicture: begin
        T := Node.Items.AddTag('hotimage', Node, Node.Root);
        WriteImage(T, TRVGraphicItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
      rvsTable: begin
        T := Node.Items.AddTag('table', Node, Node.Root);
        WriteTable(T, TRVTableItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
      rvsListMarker: begin
        T := Node.Items.AddTag('marker', Node, Node.Root);
        WriteMarker(T, TRVMarkerItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
      rvsTab: begin
        T := Node.Items.AddTag('tab', Node, Node.Root);
        WriteTab(T, TRVTabItemInfo(RVData.GetItem(SelEnd)), SelEnd); end;
      else begin
        if Assigned(OnUnknownItem) then
          OnUnknownItem(RVData, SelEnd, TCustomRVItemInfo(RVData.GetItem(SelEnd)),
            Node.Items.AddTag('', Node, Node.Root));
      end;
    end;
  end;
end;

{--------------------------------LoadStyles---------------------------------}

procedure LoadFontFromXML(Node: TXMLTag; Font: TFontInfo);
var
  I, J: Integer;
  S: string;

  procedure LoadProtection(Attr: TXMLAttr; Option: TRVProtectOption);
  begin
    if Attr.Value = '1' then
      Font.Protection := Font.Protection + [Option]
    else if Attr.Value = '0' then
      Font.Protection := Font.Protection - [Option];
  end;

begin
  for I := 0 to Node.Attr.Count - 1 do
  begin

    S := LowerCase(Node.Attr[I].Name);

    if S = 'backcolor' then
      Font.BackColor := StringToColor(Node.Attr[I].Value)

    else if S = 'charscale' then
      Font.CharScale := StrToIntDef(Node.Attr[I].Value, Font.CharScale)

    else if S = 'charspacing' then
      Font.CharSpacing := StrToIntDef(Node.Attr[I].Value, Font.CharSpacing)

    else if S = 'charset' then
    begin
      if IdentToCharset(Node.Attr[I].Value, J) then
        Font.Charset := J;
    end

    else if S = 'color' then
      Font.Color := StringToColor(Node.Attr[I].Value)

    else if S = 'fontname' then
      Font.FontName := Node.Attr[I].Value

    else if S = 'hoverbackcolor' then
      Font.HoverBackColor := StringToColor(Node.Attr[I].Value)

    else if S = 'hovercolor' then
      Font.HoverColor := StringToColor(Node.Attr[I].Value)

    else if S = 'jump' then
    begin
      if Node.Attr[I].Value = '1' then
        Font.Jump := True
      else if Node.Attr[I].Value = '0' then
        Font.Jump := False;
    end

    else if S = 'nextstyleno' then
      Font.NextStyleNo := StrToIntDef(Node.Attr[I].Value, Font.NextStyleNo)

    else if S = 'stylechangeprotect' then
      LoadProtection(Node.Attr[I], rvprStyleProtect)

    else if S = 'texteditingprotect' then
      LoadProtection(Node.Attr[I], rvprModifyProtect)

    else if S = 'textdeleteprotect' then
      LoadProtection(Node.Attr[I], rvprDeleteProtect)

    else if S = 'textconcateprotect' then
      LoadProtection(Node.Attr[I], rvprConcateProtect)

    else if S = 'styleautoswitchprotect' then
      LoadProtection(Node.Attr[I], rvprDoNotAutoSwitch)

    else if S = 'parastartprotect' then
      LoadProtection(Node.Attr[I], rvprParaStartProtect)

    else if S = 'sticking' then
      LoadProtection(Node.Attr[I], rvprSticking)

    else if S = 'sticktotop' then
      LoadProtection(Node.Attr[I], rvprStickToTop)

    else if S = 'sticktobottom' then
      LoadProtection(Node.Attr[I], rvprStickToBottom)

    else if S = 'size' then
      Font.Size := StrToIntDef(Node.Attr[I].Value, Font.Size)

    else if S = 'standard' then
    begin
      if Node.Attr[I].Value = '1' then
        Font.Standard := True
      else if Node.Attr[I].Value = '0' then
        Font.Standard := False;
    end

    else if S = 'bold' then
    begin
      if Node.Attr[I].Value = '1' then
        Font.Style := Font.Style + [fsBold]
      else if Node.Attr[I].Value = '0' then
        Font.Style := Font.Style - [fsBold];
    end

    else if S = 'italic' then
    begin
      if Node.Attr[I].Value = '1' then
        Font.Style := Font.Style + [fsItalic]
      else if Node.Attr[I].Value = '0' then
        Font.Style := Font.Style - [fsItalic];
    end

    else if S = 'underline' then
    begin
      if Node.Attr[I].Value = '1' then
        Font.Style := Font.Style + [fsUnderline]
      else if Node.Attr[I].Value = '0' then
        Font.Style := Font.Style - [fsUnderline];
    end

    else if S = 'strikeout' then
    begin
      if Node.Attr[I].Value = '1' then
        Font.Style := Font.Style + [fsStrikeOut]
      else if Node.Attr[I].Value = '0' then
        Font.Style := Font.Style - [fsStrikeOut];
    end

    else if S = 'overline' then
    begin
      if Node.Attr[I].Value = '1' then
        Font.StyleEx := Font.StyleEx + [rvfsOverline]
      else if Node.Attr[I].Value = '0' then
        Font.StyleEx := Font.StyleEx - [rvfsOverline];
    end

    else if S = 'allcaps' then
    begin
      if Node.Attr[I].Value = '1' then
        Font.StyleEx := Font.StyleEx + [rvfsAllCaps]
      else if Node.Attr[I].Value = '0' then
        Font.StyleEx := Font.StyleEx - [rvfsAllCaps];
    end

    else if S = 'name' then
      Font.StyleName := Node.Attr[I].Value

    else if S = 'unicode' then
    begin
      if Node.Attr[I].Value = '1' then
        Font.Unicode := True
      else if Node.Attr[I].Value = '0' then
        Font.Unicode := False;
    end

    else if S = 'vshift' then
      Font.VShift := StrToIntDef(Node.Attr[I].Value, Font.VShift)

    else if S = 'bidimode' then
    begin
      if LowerCase(Node.Attr[I].Value) = 'none' then
        Font.BiDiMode := rvbdUnspecified
      else if LowerCase(Node.Attr[I].Value) = 'ltr' then
        Font.BiDiMode := rvbdLeftToRight
      else if LowerCase(Node.Attr[I].Value) = 'rtl' then
        Font.BiDiMode := rvbdRightToLeft;
    end

    else if S = 'rtfcode' then
    begin
      if Node.Attr[I].Value = '1' then
        Font.Options := Font.Options + [rvteoRTFCode]
      else if Node.Attr[I].Value = '0' then
        Font.Options := Font.Options - [rvteoRTFCode]
    end

    else if S = 'htmlcode' then
    begin
      if Node.Attr[I].Value = '1' then
        Font.Options := Font.Options + [rvteoHTMLCode]
      else if Node.Attr[I].Value = '0' then
        Font.Options := Font.Options - [rvteoHTMLCode]
    end;

  end;
end;

procedure LoadParaFromXML(Node: TXMLTag; Para: TParaInfo);
var
  I: Integer;
  S, SV: string;
  {$IFNDEF RVDONOTUSETABS}
  Tab: TRVTabInfo;
  T: TXMLTag;
  J: Integer;
  {$ENDIF}
begin

  for I := 0 to Node.Items.Count - 1 do
    if LowerCase(Node.Items[I].Name) = 'tab' then
    begin
      T := Node.Items[I];
      Tab := Para.Tabs.Add;
      for J := 0 to T.Attr.Count - 1 do
      begin
        S := LowerCase(T.Attr[J].Name);
        if S = 'position' then
          Tab.Position := StrToInt(T.Attr[J].Value)
        else if S = 'align' then
        begin
          SV := LowerCase(T.Attr[J].Value);
          if SV = 'left' then
            Tab.Align := rvtaLeft
          else if SV = 'right' then
            Tab.Align := rvtaRight
          else if SV = 'center' then
            Tab.Align := rvtaCenter;
        end
        else if S = 'leader' then
          Tab.Leader := T.Attr[J].Value;
      end;
    end;

  for I := 0 to Node.Attr.Count - 1 do
  begin

    S := LowerCase(Node.Attr[I].Name);

    if S = 'align' then
    begin
      SV := LowerCase(Node.Attr[I].Value);
      if SV = 'left' then
        Para.Alignment := rvaLeft
      else if SV = 'right' then
        Para.Alignment := rvaRight
      else if SV = 'center' then
        Para.Alignment := rvaCenter
      else if SV = 'justify' then
        Para.Alignment := rvaJustify;
    end

    else if S = 'bidimode' then
    begin
      SV := LowerCase(Node.Attr[I].Value);
      if SV = 'none' then
        Para.BiDiMode := rvbdUnspecified
      else if SV = 'ltr' then
        Para.BiDiMode := rvbdLeftToRight
      else if SV = 'rtl' then
        Para.BiDiMode := rvbdRightToLeft;
    end

    else if S = 'backleftoffset' then
      Para.Background.BorderOffsets.Left := StrToIntDef(Node.Attr[I].Value, Para.Background.BorderOffsets.Left)

    else if S = 'backrightoffset' then
      Para.Background.BorderOffsets.Right := StrToIntDef(Node.Attr[I].Value, Para.Background.BorderOffsets.Right)

    else if S = 'backtopoffset' then
      Para.Background.BorderOffsets.Top := StrToIntDef(Node.Attr[I].Value, Para.Background.BorderOffsets.Top)

    else if S = 'backbottomoffset' then
      Para.Background.BorderOffsets.Bottom := StrToIntDef(Node.Attr[I].Value, Para.Background.BorderOffsets.Bottom)

    else if S = 'backcolor' then
      Para.Background.Color := StringToColor(Node.Attr[I].Value)

    else if S = 'borderleftoffset' then
      Para.Border.BorderOffsets.Left := StrToIntDef(Node.Attr[I].Value, Para.Border.BorderOffsets.Left)

    else if S = 'borderrightoffset' then
      Para.Border.BorderOffsets.Right := StrToIntDef(Node.Attr[I].Value, Para.Border.BorderOffsets.Right)

    else if S = 'bordertopoffset' then
      Para.Border.BorderOffsets.Top := StrToIntDef(Node.Attr[I].Value, Para.Border.BorderOffsets.Top)

    else if S = 'borderbottomoffset' then
      Para.Border.BorderOffsets.Bottom := StrToIntDef(Node.Attr[I].Value, Para.Border.BorderOffsets.Bottom)

    else if S = 'bordercolor' then
      Para.Border.Color := StringToColor(Node.Attr[I].Value)

    else if S = 'borderintwidth' then
      Para.Border.InternalWidth := StrToIntDef(Node.Attr[I].Value, Para.Border.InternalWidth)

    else if S = 'borderstyle' then
    begin
      SV := LowerCase(Node.Attr[I].Value);
      if SV = 'none' then
        Para.Border.Style := rvbNone
      else if SV = 'single' then
        Para.Border.Style := rvbSingle
      else if SV = 'double' then
        Para.Border.Style := rvbDouble
      else if SV = 'triple' then
        Para.Border.Style := rvbTriple
      else if SV = 'thickinside' then
        Para.Border.Style := rvbThickInside
      else if SV = 'thickoutside' then
        Para.Border.Style := rvbThickOutside;
    end

    else if S = 'borderleftvisible' then
    begin
      if Node.Attr[I].Value = '1' then
        Para.Border.VisibleBorders.Left := True
      else if Node.Attr[I].Value = '0' then
        Para.Border.VisibleBorders.Left := False;
    end

    else if S = 'borderrightvisible' then
    begin
      if Node.Attr[I].Value = '1' then
        Para.Border.VisibleBorders.Right := True
      else if Node.Attr[I].Value = '0' then
        Para.Border.VisibleBorders.Right := False;
    end

    else if S = 'bordertopvisible' then
    begin
      if Node.Attr[I].Value = '1' then
        Para.Border.VisibleBorders.Top := True
      else if Node.Attr[I].Value = '0' then
        Para.Border.VisibleBorders.Top := False;
    end

    else if S = 'borderbottomvisible' then
    begin
      if Node.Attr[I].Value = '1' then
        Para.Border.VisibleBorders.Bottom := True
      else if Node.Attr[I].Value = '0' then
        Para.Border.VisibleBorders.Bottom := False;
    end

    else if S = 'borderwidth' then
      Para.Border.Width := StrToIntDef(Node.Attr[I].Value, Para.Border.Width)

    else if S = 'firstindent' then
      Para.FirstIndent := StrToIntDef(Node.Attr[I].Value, Para.FirstIndent)

    else if S = 'linespacing' then
      Para.LineSpacing := StrToIntDef(Node.Attr[I].Value, Para.LineSpacing)

    else if S = 'linespacingtype' then
    begin
      if Node.Attr[I].Value = '%' then
        Para.LineSpacingType := rvlsPercent
      else if LowerCase(Node.Attr[I].Value) = 'spacebetween' then
        Para.LineSpacingType := rvlsSpaceBetween;
    end

    else if S = 'leftindent' then
      Para.LeftIndent := StrToIntDef(Node.Attr[I].Value, Para.LeftIndent)

    else if S = 'nextparano' then
      Para.NextParaNo := StrToIntDef(Node.Attr[I].Value, Para.NextParaNo)

    else if S = 'standard' then
    begin
      if Node.Attr[I].Value = '1' then
        Para.Standard := True
      else if Node.Attr[I].Value = '0' then
        Para.Standard := False;
    end

    else if S = 'wrap' then
    begin
      if Node.Attr[I].Value = '1' then
        Para.Options := Para.Options - [rvpaoNoWrap]
      else if Node.Attr[I].Value = '0' then
        Para.Options := Para.Options + [rvpaoNoWrap];
    end

    else if S = 'readonly' then
    begin
      if Node.Attr[I].Value = '1' then
        Para.Options := Para.Options + [rvpaoReadOnly]
      else if Node.Attr[I].Value = '0' then
        Para.Options := Para.Options - [rvpaoReadOnly];
    end

    else if S = 'styleprotect' then
    begin
      if Node.Attr[I].Value = '1' then
        Para.Options := Para.Options + [rvpaoStyleProtect]
      else if Node.Attr[I].Value = '0' then
        Para.Options := Para.Options - [rvpaoStyleProtect];
    end

    else if S = 'ignoreenter' then
    begin
      if Node.Attr[I].Value = '1' then
        Para.Options := Para.Options + [rvpaoDoNotWantReturns]
      else if Node.Attr[I].Value = '0' then
        Para.Options := Para.Options - [rvpaoDoNotWantReturns];
    end

    else if S = 'keeplinestogether' then
    begin
      if Node.Attr[I].Value = '1' then
        Para.Options := Para.Options + [rvpaoKeepLinesTogether]
      else if Node.Attr[I].Value = '0' then
        Para.Options := Para.Options - [rvpaoKeepLinesTogether];
    end

    else if S = 'keepwithnext' then
    begin
      if Node.Attr[I].Value = '1' then
        Para.Options := Para.Options + [rvpaoKeepWithNext]
      else if Node.Attr[I].Value = '0' then
        Para.Options := Para.Options - [rvpaoKeepWithNext];
    end

    else if S = 'rightindent' then
      Para.RightIndent := StrToIntDef(Node.Attr[I].Value, Para.RightIndent)

    else if S = 'spaceafter' then
      Para.SpaceAfter := StrToIntDef(Node.Attr[I].Value, Para.SpaceAfter)

    else if S = 'spacebefore' then
      Para.SpaceBefore := StrToIntDef(Node.Attr[I].Value, Para.SpaceBefore)

    else if S = 'name' then
      Para.StyleName := Node.Attr[I].Value;

  end;
end;

procedure LoadMarkerFromXML(Node: TXMLTag; Item: TRVListLevel);
var
  S: string;
  WS: WideString;
  I, J: Integer;
  Stream: TMemoryStream;
  Cl: TClass;
  B: TBitmap;
begin
  Stream := nil;
  Cl := nil;
  for I := 0 to Node.Attr.Count - 1 do
    with Node.Attr[I] do
    begin
      S := LowerCase(Name);

      if S = 'type' then
      begin
        S := LowerCase(Value);
        if S = 'bullet' then
          Item.ListType := rvlstBullet
        else if S = 'unibullet' then
          Item.ListType := rvlstUnicodeBullet
        else if S = 'picture' then
          Item.ListType := rvlstPicture
        else if S = 'imglist' then
          Item.ListType :=  rvlstImageList
        else if S = 'imglistcounter' then
          Item.ListType :=  rvlstImageListCounter
        else if S = 'decimal' then
          Item.ListType := rvlstDecimal
        else if S = 'loweralpha' then
          Item.ListType := rvlstLowerAlpha
        else if S = 'upperalpha' then
          Item.ListType := rvlstUpperAlpha
        else if S = 'lowerroman' then
          Item.ListType := rvlstLowerRoman
        else if S = 'upperroman' then
          Item.ListType := rvlstUpperRoman;
      end

      else if S = 'fontcharset' then
      begin
        if IdentToCharset(Value, J) then
          Item.Font.Charset := J;
      end

      else if S = 'fontcolor' then
        Item.Font.Color := StringToColor(Value)

      else if S = 'fontname' then
        Item.Font.Name := Value

      else if S = 'fontsize' then
        Item.Font.Size := StrToIntDef(Value, Item.Font.Size)

      else if S = 'fontbold' then
      begin
        if Value = '0' then
          Item.Font.Style := Item.Font.Style - [fsBold]
        else if Value = '1' then
          Item.Font.Style := Item.Font.Style + [fsBold];
      end

      else if S = 'fontitalic' then
      begin
        if Value = '0' then
          Item.Font.Style := Item.Font.Style - [fsItalic]
        else if Value = '1' then
          Item.Font.Style := Item.Font.Style + [fsItalic];
      end

      else if S = 'fontunderline' then
      begin
        if Value = '0' then
          Item.Font.Style := Item.Font.Style - [fsUnderline]
        else if Value = '1' then
          Item.Font.Style := Item.Font.Style + [fsUnderline];
      end

      else if S = 'fontstrikeout' then
      begin
        if Value = '0' then
          Item.Font.Style := Item.Font.Style - [fsStrikeout]
        else if Value = '1' then
          Item.Font.Style := Item.Font.Style + [fsStrikeout];
      end

      else if S = 'formatstr' then
        Item.FormatString := Value

      else if S = 'formatstrw' then begin
          if Stream <> nil then
            Stream.Clear
          else
            Stream := TMemoryStream.Create;
          RVFTextString2Stream(Value, Stream);
          Stream.Position := 0;
          SetLength(WS, Stream.Size div 2);
          Stream.ReadBuffer(Pointer(WS)^, Length(WS)*2);
          Item.FormatStringW := WS;
        end

      else if S = 'leftindent' then
        Item.LeftIndent := StrToIntDef(Value, Item.LeftIndent)

      else if S = 'firstindent' then
        Item.FirstIndent := StrToIntDef(Value, Item.FirstIndent)

      else if S = 'markerindent' then
        Item.MarkerIndent := StrToIntDef(Value, Item.MarkerIndent)

      else if S = 'markeralign' then
      begin
        S := LowerCase(Value);
        if S = 'left' then
          Item.MarkerAlignment := rvmaLeft
        else if S = 'right' then
          Item.MarkerAlignment := rvmaRight
        else if S = 'center' then
          Item.MarkerAlignment := rvmaCenter;
      end

      else if S = 'continuous' then
      begin
        if Value = '0' then
          Item.Options := Item.Options - [rvloContinuous]
        else if Value = '1' then
          Item.Options := Item.Options + [rvloContinuous];
      end

      else if S = 'levelreset' then
      begin
        if Value = '0' then
          Item.Options := Item.Options - [rvloLevelReset]
        else if Value = '1' then
          Item.Options := Item.Options + [rvloLevelReset];
      end

      else if S = 'legalstylenumbering' then
      begin
        if Value = '0' then
          Item.Options := Item.Options - [rvloLegalStyleNumbering]
        else if Value = '1' then
          Item.Options := Item.Options + [rvloLegalStyleNumbering];
      end

      else if S = 'imgsource' then
      begin
        if FileExists(Node.Root.BaseFilePath + Value) then
        begin
          Cl := nil;
          Item.Picture.LoadFromFile(Node.Root.BaseFilePath + Value);
        end;
      end

      else if S = 'img' then
      begin
        if Stream <> nil then
          Stream.Clear
        else
          Stream := TMemoryStream.Create;
        RVFTextString2Stream(Value, Stream);
      end

      else if S = 'class' then
        Cl := FindClass(Value);

    end;

  if (Stream <> nil) and (Cl <> nil) then
  begin
    Stream.Position := 0;
    Item.Picture.Graphic := RV_CreateGraphics(TGraphicClass(Cl));
    Item.Picture.Graphic.LoadFromStream(Stream);
  end;

  if Stream <> nil then
    Stream.Free;

  if Item.ListType in [rvlstImageList,rvlstImageListCounter] then
  begin
    Item.ImageIndex := 0;
    for I := 0 to Node.Items.Count - 1 do
      if LowerCase(Node.Items[I].Name) = 'img' then
      begin
        B := TBitmap.Create;
        if Node.Items[I].Value <> '' then
        begin
          Stream := TMemoryStream.Create;
          RVFTextString2Stream(Node.Items[I].Value, Stream);
          Stream.Position := 0;
          try
            B.LoadFromStream(Stream);
          finally
            Stream.Free;
          end;
        end else
        begin
          for J := 0 to Node.Items[I].Attr.Count - 1 do
            if LowerCase(Node.Items[I].Attr[J].Name) = 'source' then
              B.LoadFromFile(Node.Root.BaseFilePath + Node.Items[I].Attr[J].Value);
        end;
        if Item.ImageList = nil then
          Item.ImageList := TImageList.CreateSize(B.Width, B.Height);
        Item.ImageList.Add(B, nil);
        B.Free;
      end;
  end;
end;

procedure LoadListFromXML(Node: TXMLTag; List: TRVListInfo);
var
  I: Integer;
  S: string;
  L: TRVListLevel;
begin
  for I := 0 to Node.Attr.Count - 1 do
    with Node.Attr[I] do
    begin
      S := LowerCase(Name);

      if S = 'standard' then
      begin
        if Value = '0' then
          List.Standard := False
        else if Value = '1' then
          List.Standard := True;
      end

      else if S = 'name' then
        List.StyleName := Value;

    end;

  for I := 0 to Node.Items.Count - 1 do
    if LowerCase(Node.Items[I].Name) = 'marker' then
    begin
      L := List.Levels.Add;
      LoadMarkerFromXML(Node.Items[I], L);
    end;
end;

procedure LoadRVStyle(Node: TXMLTag; RVStyle: TRVStyle; StyleLoadingMode: TStyleLoadingMode;
  StyleMap, ParaMap, ListMap: TRVIntegerList);
var
  I: Integer;
  FontInfos: TFontInfos;
  ParaInfos: TParaInfos;
  ListInfos: TRVListInfos;
  Font: TFontInfo;
  Para: TParaInfo;
  List: TRVListInfo;
begin
  StyleMap.Clear;
  ParaMap.Clear;
  ListMap.Clear;
  if (StyleLoadingMode = slmIgnore) or (Node = nil) then
  begin
    for I := 0 to RVStyle.TextStyles.Count - 1 do
      StyleMap.Add(I);
    for I := 0 to RVStyle.ParaStyles.Count - 1 do
      ParaMap.Add(I);
    for I := 0 to RVStyle.ListStyles.Count - 1 do
      ListMap.Add(I);

    Exit;
  end;

  for I := 0 to Node.Attr.Count - 1 do
  begin
    if LowerCase(Node.Attr[I].Name) = 'inactiveselcolor' then
      RVStyle.InactiveSelColor := StringToColor(Node.Attr[I].Value)
    else if LowerCase(Node.Attr[I].Name) = 'inactiveseltextcolor' then
      RVStyle.InactiveSelTextColor := StringToColor(Node.Attr[I].Value);
  end;

  FontInfos := TFontInfos.Create(TFontInfo, nil);
  ParaInfos := TParaInfos.Create(TParaInfo, nil);
  ListInfos := TRVListInfos.Create(TRVListInfo, nil);
  try
    for I := 0 to Node.Items.Count - 1 do
    begin
      if LowerCase(Node.Items[I].Name) = 'font' then
      begin
        Font := FontInfos.Add;
        LoadFontFromXML(Node.Items[I], Font);
      end else if LowerCase(Node.Items[I].Name) = 'para' then
      begin
        Para := ParaInfos.Add;
        LoadParaFromXML(Node.Items[I], Para);
      end else if LowerCase(Node.Items[I].Name) = 'list' then
      begin
        List := ListInfos.Add;
        LoadListFromXML(Node.Items[I], List);
      end;
    end;

    if StyleLoadingMode = slmMerge then
    begin
      RVStyle.TextStyles.MergeWith(FontInfos, rvs_merge_SmartMerge, StyleMap, StyleMap);
      RVStyle.ParaStyles.MergeWith(ParaInfos, rvs_merge_SmartMerge, ParaMap, StyleMap);
      RVStyle.ListStyles.MergeWith(ListInfos, rvs_merge_SmartMerge, ListMap, StyleMap);
      Exit;
    end;

  { Bug in RichView (TRVStyle.MergeWith) }
    if StyleLoadingMode = slmMap then
    begin
      RVStyle.TextStyles.MergeWith(FontInfos, rvs_merge_Map, StyleMap, StyleMap);
      RVStyle.ParaStyles.MergeWith(ParaInfos, rvs_merge_Map, ParaMap, StyleMap);
      RVStyle.ListStyles.MergeWith(ListInfos, rvs_merge_Map, ListMap, StyleMap);
      Exit;
    end;

    RVStyle.TextStyles := FontInfos;
    RVStyle.ParaStyles := ParaInfos;
    RVStyle.ListStyles := ListInfos;

    for I := 0 to RVStyle.TextStyles.Count - 1 do
      StyleMap.Add(I);
    for I := 0 to RVStyle.ParaStyles.Count - 1 do
      ParaMap.Add(I);
    for I := 0 to RVStyle.ListStyles.Count - 1 do
      ListMap.Add(I);
  finally
    FontInfos.Free;
    ParaInfos.Free;
    ListInfos.Free;
  end;
end;

{--------------------------------SaveStyles---------------------------------}

procedure SaveFontToXML(Node: TXMLTag; Font: TFontInfo; SaveStyleName: Boolean);
var
  S: string;
begin

  if Font.BackColor <> clNone then
    Node.Attr.AddAttr('backcolor', ColorToString(Font.BackColor));

  if Font.CharScale <> 100 then
    Node.Attr.AddAttr('charscale', IntToStr(Font.CharScale));

  if Font.CharSpacing <> 0 then
    Node.Attr.AddAttr('charspacing', ColorToString(Font.CharSpacing));

  if Font.Charset <> DEFAULT_CHARSET then
    if CharsetToIdent(Font.Charset, S) then
      Node.Attr.AddAttr('charset', S);

  if Font.Color <> clWindowText then
    Node.Attr.AddAttr('color', ColorToString(Font.Color));

  if Font.FontName <> 'Arial' then
    Node.Attr.AddAttr('fontname', Font.FontName);

  if Font.Jump then
  begin
    Node.Attr.AddAttr('jump', '1');

    if Font.HoverBackColor <> clNone then
      Node.Attr.AddAttr('hoverbackcolor', ColorToString(Font.HoverBackColor));

    if Font.HoverColor <> clNone then
      Node.Attr.AddAttr('hovercolor', ColorToString(Font.HoverColor));

  end;

  if Font.NextStyleNo <> -1 then
    Node.Attr.AddAttr('nextstyleno', IntToStr(Font.NextStyleNo));

  if rvprStyleProtect in Font.Protection then
    Node.Attr.AddAttr('stylechangeprotect', '1');

  if rvprModifyProtect in Font.Protection then
    Node.Attr.AddAttr('texteditingprotect', '1');

  if rvprDeleteProtect in Font.Protection then
    Node.Attr.AddAttr('textdeleteprotect', '1');

  if rvprConcateProtect in Font.Protection then
    Node.Attr.AddAttr('textconcateprotect', '1');

  if rvprDoNotAutoSwitch in Font.Protection then
    Node.Attr.AddAttr('styleautoswitchprotect', '1');

  if rvprParaStartProtect in Font.Protection then
    Node.Attr.AddAttr('parastartprotect', '1');

  if rvprSticking in Font.Protection then
    Node.Attr.AddAttr('sticking', '1');

  if rvprStickToTop in Font.Protection then
    Node.Attr.AddAttr('sticktotop', '1');

  if rvprStickToBottom in Font.Protection then
    Node.Attr.AddAttr('sticktobottom', '1');

  if Font.Size <> 10 then
    Node.Attr.AddAttr('size', IntToStr(Font.Size));

  if not Font.Standard then
    Node.Attr.AddAttr('standard', '0');

  if fsBold in Font.Style then
    Node.Attr.AddAttr('bold', '1');

  if fsItalic in Font.Style then
    Node.Attr.AddAttr('italic', '1');

  if fsUnderline in Font.Style then
    Node.Attr.AddAttr('underline', '1');

  if fsStrikeOut in Font.Style then
    Node.Attr.AddAttr('strikeout', '1');

  if rvfsOverline in Font.StyleEx then
    Node.Attr.AddAttr('overline', '1');

  if rvfsAllCaps in Font.StyleEx then
    Node.Attr.AddAttr('allcaps', '1');

  if SaveStyleName then
    if Font.StyleName <> 'Font Style' then
      Node.Attr.AddAttr('name', Font.StyleName);

  if Font.Unicode then
    Node.Attr.AddAttr('unicode', '1');

  if Font.VShift <> 0 then
    Node.Attr.AddAttr('vshift', IntToStr(Font.VShift));

  if Font.BiDiMode <> rvbdUnspecified then
  begin
    if Font.BiDiMode = rvbdLeftToRight then
      Node.Attr.AddAttr('bidimode', 'ltr')
    else if Font.BiDiMode = rvbdRightToLeft then
      Node.Attr.AddAttr('bidimode', 'rtl');
  end;

  if rvteoRTFCode in Font.Options then
    Node.Attr.AddAttr('rtfcode', '1');

  if rvteoHTMLCode in Font.Options then
    Node.Attr.AddAttr('htmlcode', '1');

end;

procedure SaveParaToXML(Node: TXMLTag; Para: TParaInfo; SaveStyleName: Boolean);
    {$IFNDEF RVDONOTUSETABS}
var i: Integer;    
    T: TXMLTag;
    {$ENDIF}
begin

  if Para.Alignment <> rvaLeft then
    if Para.Alignment = rvaRight then
      Node.Attr.AddAttr('align', 'right')
    else if Para.Alignment = rvaCenter then
      Node.Attr.AddAttr('align', 'center')
    else if Para.Alignment = rvaJustify then
      Node.Attr.AddAttr('align', 'justify');

  if Para.BiDiMode <> rvbdUnspecified then
  begin
    if Para.BiDiMode = rvbdLeftToRight then
      Node.Attr.AddAttr('bidimode', 'ltr')
    else if Para.BiDiMode = rvbdRightToLeft then
      Node.Attr.AddAttr('bidimode', 'rtl');
  end;

  if Para.Background.BorderOffsets.Left <> 0 then
    Node.Attr.AddAttr('backleftoffset', IntToStr(Para.Background.BorderOffsets.Left));

  if Para.Background.BorderOffsets.Right <> 0 then
    Node.Attr.AddAttr('backrightoffset', IntToStr(Para.Background.BorderOffsets.Right));

  if Para.Background.BorderOffsets.Top <> 0 then
    Node.Attr.AddAttr('backtopoffset', IntToStr(Para.Background.BorderOffsets.Top));

  if Para.Background.BorderOffsets.Bottom <> 0 then
    Node.Attr.AddAttr('backbottomoffset', IntToStr(Para.Background.BorderOffsets.Bottom));

  if Para.Background.Color <> clNone then
    Node.Attr.AddAttr('backcolor', ColorToString(Para.Background.Color));

  if Para.Border.BorderOffsets.Left <> 0 then
    Node.Attr.AddAttr('borderleftoffset', IntToStr(Para.Border.BorderOffsets.Left));

  if Para.Border.BorderOffsets.Right <> 0 then
    Node.Attr.AddAttr('borderrightoffset', IntToStr(Para.Border.BorderOffsets.Right));

  if Para.Border.BorderOffsets.Top <> 0 then
    Node.Attr.AddAttr('bordertopoffset', IntToStr(Para.Border.BorderOffsets.Top));

  if Para.Border.BorderOffsets.Bottom <> 0 then
    Node.Attr.AddAttr('borderbottomoffset', IntToStr(Para.Border.BorderOffsets.Bottom));

  if Para.Border.Color <> clWindowText then
    Node.Attr.AddAttr('bordercolor', ColorToString(Para.Border.Color));

  if Para.Border.InternalWidth <> 1 then
    Node.Attr.AddAttr('borderintwidth', IntToStr(Para.Border.InternalWidth));

  if Para.Border.Style <> rvbNone then
    if Para.Border.Style = rvbSingle then
      Node.Attr.AddAttr('borderstyle', 'single')
    else if Para.Border.Style = rvbDouble then
      Node.Attr.AddAttr('borderstyle', 'double')
    else if Para.Border.Style = rvbTriple then
      Node.Attr.AddAttr('borderstyle', 'triple')
    else if Para.Border.Style = rvbThickInside then
      Node.Attr.AddAttr('borderstyle', 'thickinside')
    else if Para.Border.Style = rvbThickOutside then
      Node.Attr.AddAttr('borderstyle', 'thickoutside');

  if not Para.Border.VisibleBorders.Left then
    Node.Attr.AddAttr('borderleftvisible', '0');

  if not Para.Border.VisibleBorders.Right then
    Node.Attr.AddAttr('borderrightvisible', '0');

  if not Para.Border.VisibleBorders.Top then
    Node.Attr.AddAttr('bordertopvisible', '0');

  if not Para.Border.VisibleBorders.Bottom then
    Node.Attr.AddAttr('borderbottomvisible', '0');

  if Para.Border.Width <> 1 then
    Node.Attr.AddAttr('borderwidth', IntToStr(Para.Border.Width));

  if Para.FirstIndent <> 0 then
    Node.Attr.AddAttr('firstindent', IntToStr(Para.FirstIndent));

  if Para.LineSpacing <> 100 then
    Node.Attr.AddAttr('linespacing', IntToStr(Para.LineSpacing));

  if Para.LineSpacingType <> rvlsPercent then
    Node.Attr.AddAttr('linespacingtype', 'spacebetween');

  if Para.LeftIndent <> 0 then
    Node.Attr.AddAttr('leftindent', IntToStr(Para.LeftIndent));

  if Para.NextParaNo <> -1 then
    Node.Attr.AddAttr('nextparano', IntToStr(Para.NextParaNo));

  if rvpaoNoWrap in Para.Options then
    Node.Attr.AddAttr('wrap', '0');

  if rvpaoReadOnly in Para.Options then
    Node.Attr.AddAttr('readonly', '1');

  if rvpaoStyleProtect in Para.Options then
    Node.Attr.AddAttr('styleprotect', '1');

  if rvpaoDoNotWantReturns in Para.Options then
    Node.Attr.AddAttr('ignoreenter', '1');

  if rvpaoKeepLinesTogether in Para.Options then
    Node.Attr.AddAttr('keeplinestogether', '1');

  if rvpaoKeepWithNext in Para.Options then
    Node.Attr.AddAttr('keepwithnext', '1');

  if Para.RightIndent <> 0 then
    Node.Attr.AddAttr('rightindent', IntToStr(Para.RightIndent));

  if Para.SpaceAfter <> 0 then
    Node.Attr.AddAttr('spaceafter', IntToStr(Para.SpaceAfter));

  if Para.SpaceBefore <> 0 then
    Node.Attr.AddAttr('spacebefore', IntToStr(Para.SpaceBefore));

  if not Para.Standard then
    Node.Attr.AddAttr('standard', '0');

  {$IFNDEF RVDONOTUSETABS}
  for i := 0 to Para.Tabs.Count-1 do
  begin
    T := Node.Items.AddTag('tab', Node, Node.Root);
    T.Attr.AddAttr('position', IntToStr(Para.Tabs[i].Position));
    case Para.Tabs[i].Align of
      rvtaRight:
        T.Attr.AddAttr('align', 'right');
      rvtaCenter:
        T.Attr.AddAttr('align', 'center');
    end;
    if Para.Tabs[i].Leader<>'' then
      T.Attr.AddAttr('leader', Para.Tabs[i].Leader);
  end;
  {$ENDIF}

  if SaveStyleName then
    if Para.StyleName <> 'Paragraph Style' then
      Node.Attr.AddAttr('name', Para.StyleName);

end;

procedure SaveMarkerToXML(Node: TXMLTag; Item: TRVListLevel; SaveImgToFiles: Boolean);

  procedure SaveMarkerFont;
  var
    S: string;
  begin
    with Node.Attr, Item.Font do
    begin

      if Charset <> DEFAULT_CHARSET then
        if CharsetToIdent(Charset, S) then
          AddAttr('fontcharset', S);

      if Color <> clWindowText then
        AddAttr('fontcolor', ColorToString(Color));

      if Name <> 'Arial' then
        AddAttr('fontname', Name);

      if Size <> 8 then
        AddAttr('fontsize', IntToStr(Size));

      if fsBold in Style then
        AddAttr('fontbold', '1');

      if fsItalic in Style then
        AddAttr('fontitalic', '1');

      if fsUnderline in Style then
        AddAttr('fontunderline', '1');

      if fsStrikeout in Style then
        AddAttr('fontstrikeout', '1');

    end;
  end;

  procedure SaveDefProps;
  var Stream: TMemoryStream;
  begin
    with Node.Attr do
    begin
      if Item.FormatString <> '' then
        AddAttr('formatstr', Item.FormatString);

      if Item.FormatStringW <> '' then begin
        Stream := TMemoryStream.Create;
        Stream.WriteBuffer(Pointer(Item.FormatStringW)^, Length(Item.FormatStringW)*2);
        Stream.Position := 0;
        AddAttr('formatstrw', RVFStream2TextString(Stream));
        Stream.Free;
      end;

      if Item.LeftIndent <> 0 then
        AddAttr('leftindent', IntToStr(Item.LeftIndent));

      if Item.FirstIndent <> 10 then
        AddAttr('firstindent', IntToStr(Item.FirstIndent));

      if Item.MarkerIndent <> 0 then
        AddAttr('markerindent', IntToStr(Item.MarkerIndent));

      if Item.MarkerAlignment <> rvmaLeft then
        if Item.MarkerAlignment = rvmaRight then
          AddAttr('markeralign', 'right')
        else
          AddAttr('markeralign', 'center');

      if not (rvloContinuous in Item.Options) then
        AddAttr('continuous', '0');

      if not (rvloLevelReset in Item.Options) then
        AddAttr('levelreset', '0');

      if rvloLegalStyleNumbering in Item.Options then
        AddAttr('legalstylenumbering', '1'); 
    end;
  end;

  procedure SavePicture(Gr: TGraphic);
  var
    Stream: TMemoryStream;
    FileName: string;
  begin
    if SaveImgToFiles then
    begin
      FileName := GetImageFileName(Node.Root.BaseFilePath, 'marker', GraphicExtension(TGraphicClass(Gr.ClassType)));
      Gr.SaveToFile(FileName);
      Node.Attr.AddAttr('imgsource', ExtractFileName(FileName));
    end else
    begin
      Stream := TMemoryStream.Create;
      try
        Gr.SaveToStream(Stream);
        Node.Attr.AddAttr('class', Gr.ClassName);
        Node.Attr.AddAttr('img', RVFStream2TextString(Stream));
      finally
        Stream.Free;
      end;
    end;
  end;

  procedure SaveImgList;
  var
    B: TBitmap;
    N: TXMLTag;
    I: Integer;
    S: TMemoryStream;
  begin
    for I := Item.ImageIndex to Item.ImageList.Count - 1 do
    begin
      N := Node.Items.AddTag('img', Node, Node.Root);
      B := TBitmap.Create;      
      try
        Item.ImageList.GetBitmap(I, B);
        if SaveImgToFiles then
          N.Attr.AddAttr('source', GetImageFileName(Node.Root.BaseFilePath, 'markeril',
            GraphicExtension(TBitmap)))
        else
        begin
          S := TMemoryStream.Create;
          B.SaveToStream(S);
          S.Position := 0;
          try
            N.Value := RVFStream2TextString(S);
          finally
            S.Free;
          end;
        end;
      finally
        B.Free;
      end;
    end;
  end;

begin
  with Node.Attr do
    case Item.ListType of

      rvlstBullet: begin
        AddAttr('type', 'bullet');
        SaveDefProps;
        SaveMarkerFont;
      end;

      rvlstUnicodeBullet: begin
        AddAttr('type', 'unibullet');
        SaveDefProps;
        SaveMarkerFont;
      end;

      rvlstPicture: begin
        AddAttr('type', 'picture');
        SaveDefProps;
        SavePicture(Item.Picture.Graphic);
      end;

      rvlstImageList: begin
        AddAttr('type', 'imglist');
        SaveDefProps;
        SaveImgList;
      end;

      rvlstImageListCounter: begin
        AddAttr('type', 'imglistcounter');
        SaveDefProps;
        SaveImgList;
      end;

      rvlstDecimal: begin
        AddAttr('type', 'decimal');
        SaveDefProps;
        SaveMarkerFont;
      end;

      rvlstLowerAlpha: begin
        AddAttr('type', 'loweralpha');
        SaveDefProps;
        SaveMarkerFont;
      end;

      rvlstUpperAlpha: begin
        AddAttr('type', 'upperalpha');
        SaveDefProps;
        SaveMarkerFont;
      end;

      rvlstLowerRoman: begin
        AddAttr('type', 'lowerroman');
        SaveDefProps;
        SaveMarkerFont;
      end;

      rvlstUpperRoman: begin
        AddAttr('type', 'upperroman');
        SaveDefProps;
        SaveMarkerFont;
      end;

    end;
end;

procedure SaveListToXML(Node: TXMLTag; Item: TRVListInfo; SaveStyleNames: Boolean; SaveImgToFiles: Boolean);
var
  I: Integer;
begin
  with Node.Attr do
  begin
    if not Item.Standard then
      AddAttr('standard', '0');

    if SaveStyleNames then
      if Item.StyleName <> 'List Style' then
        AddAttr('name', Item.StyleName);

    for I := 0 to Item.Levels.Count - 1 do
      SaveMarkerToXML(Node.Items.AddTag('marker', Node, Node.Root), Item.Levels[I], SaveImgToFiles);
  end;
end;

procedure SaveRVStyle(Node: TXMLTag; RVStyle: TRVStyle; SaveStyleNames: Boolean; SaveImgToFiles: Boolean);
var
  I: Integer;
  T: TXMLTag;
begin  
  Node.Attr.AddAttr('inactiveselcolor', ColorToString(RVStyle.InactiveSelColor));
  Node.Attr.AddAttr('inactiveseltextcolor', ColorToString(RVStyle.InactiveSelTextColor));

  for I := 0 to RVStyle.TextStyles.Count - 1 do
  begin
    T := Node.Items.AddTag('font', Node, Node.Root);
    SaveFontToXML(T, RVStyle.TextStyles[I], SaveStyleNames);
  end;

  for I := 0 to RVStyle.ParaStyles.Count - 1 do
  begin
    T := Node.Items.AddTag('para', Node, Node.Root);
    SaveParaToXML(T, RVStyle.ParaStyles[I], SaveStyleNames);
  end;
  
  for I := 0 to RVStyle.ListStyles.Count - 1 do
  begin
    T := Node.Items.AddTag('list', Node, Node.Root);
    SaveListToXML(T, RVStyle.ListStyles[I], SaveStyleNames, SaveImgToFiles);
  end;
end;

{---------------------------Save/Load Background----------------------------}

procedure LoadBackgroundFromXML(Node: TXMLTag; RichView: TCustomRichView);
var
  Stream: TMemoryStream;
  I: Integer;
  S: string;
begin
  for I := 0 to Node.Attr.Count - 1 do
  begin
    S := LowerCase(Node.Attr[I].Name);

    if S = 'backstyle' then
    begin
      if LowerCase(Node.Attr[I].Value) = 'color' then
        RichView.BackgroundStyle := bsNoBitmap
      else if LowerCase(Node.Attr[I].Value) = 'stretch' then
        RichView.BackgroundStyle := bsStretched
      else if LowerCase(Node.Attr[I].Value) = 'tile' then
        RichView.BackgroundStyle := bsTiledAndScrolled
      else if LowerCase(Node.Attr[I].Value) = 'fixedtile' then
        RichView.BackgroundStyle := bsTiled
      else if LowerCase(Node.Attr[I].Value) = 'center' then
        RichView.BackgroundStyle := bsCentered;
    end

    else if S = 'backcolor' then
      RichView.Color := StringToColor(Node.Attr[I].Value)

    else if S = 'backimg' then
    begin
      Stream := TMemoryStream.Create;
      try
        RVFTextString2Stream(Node.Attr[I].Value, Stream);
        Stream.Position := 0;
        RichView.BackgroundBitmap.LoadFromStream(Stream);
      finally
        Stream.Free;
      end;
    end

    else if S = 'backimgsource' then
      RichView.BackgroundBitmap.LoadFromFile(Node.Root.BaseFilePath + Node.Attr[I].Value)

    else if S = 'hscrollbarvisible' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.HScrollVisible := True
      else if Node.Attr[I].Value = '0' then
        RichView.HScrollVisible := False;
    end

    else if S = 'vscrollbarvisible' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.VScrollVisible := True
      else if Node.Attr[I].Value = '0' then
        RichView.VScrollVisible := False;
    end

    else if S = 'vsmallstep' then
      RichView.VSmallStep := StrToIntDef(Node.Attr[I].Value, RichView.VSmallStep)

    else if S = 'maxtextwidth' then
      RichView.MaxTextWidth := StrToIntDef(Node.Attr[I].Value, RichView.MaxTextWidth)

    else if S = 'mintextwidth' then
      RichView.MinTextWidth := StrToIntDef(Node.Attr[I].Value, RichView.MinTextWidth)

    else if S = 'leftmargin' then
      RichView.LeftMargin := StrToIntDef(Node.Attr[I].Value, RichView.LeftMargin)

    else if S = 'rightmargin' then
      RichView.RightMargin := StrToIntDef(Node.Attr[I].Value, RichView.RightMargin)

    else if S = 'topmargin' then
      RichView.TopMargin := StrToIntDef(Node.Attr[I].Value, RichView.TopMargin)

    else if S = 'bottommargin' then
      RichView.BottomMargin := StrToIntDef(Node.Attr[I].Value, RichView.BottomMargin)

    else if S = 'scrolltoend' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.Options := RichView.Options + [rvoScrollToEnd]
      else if Node.Attr[I].Value = '0' then
        RichView.Options := RichView.Options - [rvoScrollToEnd]
    end

    else if S = 'singleclick' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.Options := RichView.Options + [rvoSingleClick]
      else if Node.Attr[I].Value = '0' then
        RichView.Options := RichView.Options - [rvoSingleClick]
    end

    else if S = 'clienttextwidth' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.Options := RichView.Options + [rvoClientTextWidth]
      else if Node.Attr[I].Value = '0' then
        RichView.Options := RichView.Options - [rvoClientTextWidth]
    end

    else if S = 'showcheckpoints' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.Options := RichView.Options + [rvoShowCheckpoints]
      else if Node.Attr[I].Value = '0' then
        RichView.Options := RichView.Options - [rvoShowCheckpoints]
    end

    else if S = 'showpagebreaks' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.Options := RichView.Options + [rvoShowPageBreaks]
      else if Node.Attr[I].Value = '0' then
        RichView.Options := RichView.Options - [rvoShowPageBreaks]
    end

    else if S = 'autocopytext' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.Options := RichView.Options + [rvoAutoCopyText]
      else if Node.Attr[I].Value = '0' then
        RichView.Options := RichView.Options - [rvoAutoCopyText]
    end

    else if S = 'autocopyunitext' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.Options := RichView.Options + [rvoAutoCopyUnicodeText]
      else if Node.Attr[I].Value = '0' then
        RichView.Options := RichView.Options - [rvoAutoCopyUnicodeText]
    end

    else if S = 'autocopyrtf' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.Options := RichView.Options + [rvoAutoCopyRTF]
      else if Node.Attr[I].Value = '0' then
        RichView.Options := RichView.Options - [rvoAutoCopyRTF]
    end

    else if S = 'autocopyrvf' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.Options := RichView.Options + [rvoAutoCopyRVF]
      else if Node.Attr[I].Value = '0' then
        RichView.Options := RichView.Options - [rvoAutoCopyRVF]
    end

    else if S = 'autocopyimage' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.Options := RichView.Options + [rvoAutoCopyImage]
      else if Node.Attr[I].Value = '0' then
        RichView.Options := RichView.Options - [rvoAutoCopyImage]
    end

    else if S = 'formatinvalidate' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.Options := RichView.Options + [rvoFormatInvalidate]
      else if Node.Attr[I].Value = '0' then
        RichView.Options := RichView.Options - [rvoFormatInvalidate]
    end

    else if S = 'dblclickselectword' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.Options := RichView.Options + [rvoDblClickSelectsWord]
      else if Node.Attr[I].Value = '0' then
        RichView.Options := RichView.Options - [rvoDblClickSelectsWord]
    end

    else if S = 'rightclickdeselect' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.Options := RichView.Options + [rvoRClickDeselects]
      else if Node.Attr[I].Value = '0' then
        RichView.Options := RichView.Options - [rvoRClickDeselects]
    end

    else if S = 'checkevent' then
    begin
      if LowerCase(Node.Attr[I].Value) = 'none' then
        RichView.CPEventKind := cpeNone
      else if LowerCase(Node.Attr[I].Value) = 'sectionstart' then
        RichView.CPEventKind := cpeAsSectionStart
      else if LowerCase(Node.Attr[I].Value) = 'visible' then
        RichView.CPEventKind := cpeWhenVisible;
    end

    else if S = 'bidimode' then
    begin
      if LowerCase(Node.Attr[I].Value) = 'none' then
        RichView.BiDiMode := rvbdUnspecified
      else if LowerCase(Node.Attr[I].Value) = 'ltr' then
        RichView.BiDiMode := rvbdLeftToRight
      else if LowerCase(Node.Attr[I].Value) = 'rtl' then
        RichView.BiDiMode := rvbdRightToLeft;
    end

    else if S = 'rtfsavestylesheet' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.RTFOptions := RichView.RTFOptions + [rvrtfSaveStyleSheet]
      else if Node.Attr[I].Value = '0' then
        RichView.RTFOptions := RichView.RTFOptions - [rvrtfSaveStyleSheet]
    end

    else if S = 'rtfunicodetoansi' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.RTFOptions := RichView.RTFOptions + [rvrtfDuplicateUnicode]
      else if Node.Attr[I].Value = '0' then
        RichView.RTFOptions := RichView.RTFOptions - [rvrtfDuplicateUnicode]
    end

    else if S = 'rtfemftowmf' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.RTFOptions := RichView.RTFOptions + [rvrtfSaveEMFAsWMF]
      else if Node.Attr[I].Value = '0' then
        RichView.RTFOptions := RichView.RTFOptions - [rvrtfSaveEMFAsWMF]
    end

    else if S = 'rtftransformjpeg' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.RTFOptions := RichView.RTFOptions - [rvrtfSaveJpegAsJpeg]
      else if Node.Attr[I].Value = '0' then
        RichView.RTFOptions := RichView.RTFOptions + [rvrtfSaveJpegAsJpeg]
    end

    else if S = 'rtfdefaultimgformat' then
    begin
      if LowerCase(Node.Attr[I].Value) = 'bmp' then
        RichView.RTFOptions := RichView.RTFOptions + [rvrtfSaveBitmapDefault]
      else if LowerCase(Node.Attr[I].Value) = 'wmf' then
        RichView.RTFOptions := RichView.RTFOptions - [rvrtfSaveBitmapDefault]
      else if LowerCase(Node.Attr[I].Value) = 'emf' then
      begin
        RichView.RTFOptions := RichView.RTFOptions - [rvrtfSaveBitmapDefault];
        RichView.RTFOptions := RichView.RTFOptions + [rvrtfSaveEMFDefault];
      end;
    end

    else if S = 'rvfautosaveimages' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.RVFOptions := RichView.RVFOptions + [rvfoSavePicturesBody]
      else if Node.Attr[I].Value = '0' then
        RichView.RVFOptions := RichView.RVFOptions - [rvfoSavePicturesBody]
    end

    else if S = 'rvfautosavecontrols' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.RVFOptions := RichView.RVFOptions + [rvfoSaveControlsBody]
      else if Node.Attr[I].Value = '0' then
        RichView.RVFOptions := RichView.RVFOptions - [rvfoSaveControlsBody]
    end

    else if S = 'rvfusestylenames' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.RVFOptions := RichView.RVFOptions + [rvfoUseStyleNames]
      else if Node.Attr[I].Value = '0' then
        RichView.RVFOptions := RichView.RVFOptions - [rvfoUseStyleNames]
    end

    else if S = 'rvfsavetextstyles' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.RVFOptions := RichView.RVFOptions + [rvfoSaveTextStyles]
      else if Node.Attr[I].Value = '0' then
        RichView.RVFOptions := RichView.RVFOptions - [rvfoSaveTextStyles]
    end

    else if S = 'rvfsaveparastyles' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.RVFOptions := RichView.RVFOptions + [rvfoSaveParaStyles]
      else if Node.Attr[I].Value = '0' then
        RichView.RVFOptions := RichView.RVFOptions - [rvfoSaveParaStyles]
    end

    else if S = 'rvfsavebinary' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.RVFOptions := RichView.RVFOptions + [rvfoSaveBinary]
      else if Node.Attr[I].Value = '0' then
        RichView.RVFOptions := RichView.RVFOptions - [rvfoSaveBinary]
    end

    else if S = 'rvfsavebackground' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.RVFOptions := RichView.RVFOptions + [rvfoSaveBack]
      else if Node.Attr[I].Value = '0' then
        RichView.RVFOptions := RichView.RVFOptions - [rvfoSaveBack]
    end

    else if S = 'rvfsavelayout' then
    begin
      if Node.Attr[I].Value = '1' then
        RichView.RVFOptions := RichView.RVFOptions + [rvfoSaveLayout]
      else if Node.Attr[I].Value = '0' then
        RichView.RVFOptions := RichView.RVFOptions - [rvfoSaveLayout];
    end;

    if not (RichView is TCustomRichViewEdit) then
    begin

      if S = 'tabnavigation' then
      begin
        if LowerCase(Node.Attr[I].Value) = 'none' then
          RichView.TabNavigation := rvtnNone
        else if LowerCase(Node.Attr[I].Value) = 'tab' then
          RichView.TabNavigation := rvtnTab
        else if LowerCase(Node.Attr[I].Value) = 'ctrltab' then
          RichView.TabNavigation := rvtnCtrlTab;
      end

      else if S = 'allowselection' then
      begin
        if Node.Attr[I].Value = '1' then
          RichView.Options := RichView.Options + [rvoAllowSelection]
        else if Node.Attr[I].Value = '0' then
          RichView.Options := RichView.Options - [rvoAllowSelection]
      end;

    end;

    if RichView is TCustomRichViewEdit then
    begin

      if S = 'cleartagbystylechange' then
      begin
        if Node.Attr[I].Value = '1' then
          TCustomRichViewEdit(RichView).EditorOptions := TCustomRichViewEdit(RichView).EditorOptions +
            [rvoClearTagOnStyleApp]
        else if Node.Attr[I].Value = '0' then
          TCustomRichViewEdit(RichView).EditorOptions := TCustomRichViewEdit(RichView).EditorOptions -
            [rvoClearTagOnStyleApp]
      end

      else if S = 'ctrljump' then
      begin
        if Node.Attr[I].Value = '1' then
          TCustomRichViewEdit(RichView).EditorOptions := TCustomRichViewEdit(RichView).EditorOptions +
            [rvoCtrlJumps]
        else if Node.Attr[I].Value = '0' then
          TCustomRichViewEdit(RichView).EditorOptions := TCustomRichViewEdit(RichView).EditorOptions -
            [rvoCtrlJumps]
      end

      else if S = 'ignoreenter' then
      begin
        if Node.Attr[I].Value = '1' then
          TCustomRichViewEdit(RichView).EditorOptions := TCustomRichViewEdit(RichView).EditorOptions +
            [rvoDoNotWantReturns]
        else if Node.Attr[I].Value = '0' then
          TCustomRichViewEdit(RichView).EditorOptions := TCustomRichViewEdit(RichView).EditorOptions -
            [rvoDoNotWantReturns]
      end

      else if S = 'ignoretab' then
      begin
        if Node.Attr[I].Value = '1' then
          TCustomRichViewEdit(RichView).EditorOptions := TCustomRichViewEdit(RichView).EditorOptions -
            [rvoWantTabs]
        else if Node.Attr[I].Value = '0' then
          TCustomRichViewEdit(RichView).EditorOptions := TCustomRichViewEdit(RichView).EditorOptions +
            [rvoWantTabs]
      end

      else if S = 'langautoswitch' then
      begin
        if Node.Attr[I].Value = '1' then
          TCustomRichViewEdit(RichView).EditorOptions := TCustomRichViewEdit(RichView).EditorOptions +
            [rvoAutoSwitchLang]
        else if Node.Attr[I].Value = '0' then
          TCustomRichViewEdit(RichView).EditorOptions := TCustomRichViewEdit(RichView).EditorOptions -
            [rvoAutoSwitchLang]
      end;

    end;
  end;
end;

procedure SaveBackgroundToXML(Node: TXMLTag; RichView: TCustomRichView; SaveImgToFile: Boolean);
var
  Stream: TMemoryStream;
begin

  if RichView.BackgroundStyle = bsNoBitmap then
    Node.Attr.AddAttr('backstyle', 'color')
  else if RichView.BackgroundStyle = bsStretched then
    Node.Attr.AddAttr('backstyle', 'stretch')
  else if RichView.BackgroundStyle = bsTiledAndScrolled then
    Node.Attr.AddAttr('backstyle', 'tile')
  else if RichView.BackgroundStyle = bsTiled then
    Node.Attr.AddAttr('backstyle', 'fixedtile')
  else if RichView.BackgroundStyle = bsCentered then
    Node.Attr.AddAttr('backstyle', 'center');

  if ((RichView.BackgroundStyle = bsNoBitmap) or (RichView.BackgroundStyle = bsCentered) or
    RichView.BackgroundBitmap.Empty) then
      Node.Attr.AddAttr('backcolor', ColorToString(RichView.Color));

  if (RichView.BackgroundStyle <> bsNoBitmap) and (not RichView.BackgroundBitmap.Empty) then
  begin
  
    if SaveImgToFile then
    begin
      RichView.BackgroundBitmap.SaveToFile(Node.Root.BaseFilePath + 'backimg.bmp');
      Node.Attr.AddAttr('backimgsource', 'backimg.bmp');
    end else begin
      Stream := TMemoryStream.Create;
      try
        RichView.BackgroundBitmap.SaveToStream(Stream);
        Stream.Position := 0;
        Node.Attr.AddAttr('backimg', RVFStream2TextString(Stream));
      finally
        Stream.Free;
      end;
    end;
    
  end;

  if RichView.HScrollVisible then
    Node.Attr.AddAttr('hscrollbarvisible', '1')
  else
    Node.Attr.AddAttr('hscrollbarvisible', '0');

  if RichView.VScrollVisible then
    Node.Attr.AddAttr('vscrollbarvisible', '1')
  else
    Node.Attr.AddAttr('vscrollbarvisible', '0');

  Node.Attr.AddAttr('vsmallstep', IntToStr(RichView.VSmallStep));

  Node.Attr.AddAttr('maxtextwidth', IntToStr(RichView.MaxTextWidth));

  Node.Attr.AddAttr('mintextwidth', IntToStr(RichView.MinTextWidth));

  Node.Attr.AddAttr('leftmargin', IntToStr(RichView.LeftMargin));

  Node.Attr.AddAttr('rightmargin', IntToStr(RichView.RightMargin));

  Node.Attr.AddAttr('topmargin', IntToStr(RichView.TopMargin));

  Node.Attr.AddAttr('bottommargin', IntToStr(RichView.BottomMargin));

  if rvoScrollToEnd in RichView.Options then
    Node.Attr.AddAttr('scrolltoend', '1')
  else
    Node.Attr.AddAttr('scrolltoend', '0');

  if rvoSingleClick in RichView.Options then
    Node.Attr.AddAttr('singleclick', '1')
  else
    Node.Attr.AddAttr('singleclick', '0');

  if rvoClientTextWidth in RichView.Options then
    Node.Attr.AddAttr('clienttextwidth', '1')
  else
    Node.Attr.AddAttr('clienttextwidth', '0');

  if rvoShowCheckpoints in RichView.Options then
    Node.Attr.AddAttr('showcheckpoints', '1')
  else
    Node.Attr.AddAttr('showcheckpoints', '0');

  if rvoShowPageBreaks in RichView.Options then
    Node.Attr.AddAttr('showpagebreaks', '1')
  else
    Node.Attr.AddAttr('showpagebreaks', '0');

  if rvoAutoCopyText in RichView.Options then
    Node.Attr.AddAttr('autocopytext', '1')
  else
    Node.Attr.AddAttr('autocopytext', '0');

  if rvoAutoCopyUnicodeText in RichView.Options then
    Node.Attr.AddAttr('autocopyunitext', '1')
  else
    Node.Attr.AddAttr('autocopyunitext', '0');

  if rvoAutoCopyRTF in RichView.Options then
    Node.Attr.AddAttr('autocopyrtf', '1')
  else
    Node.Attr.AddAttr('autocopyrtf', '0');

  if rvoAutoCopyRVF in RichView.Options then
    Node.Attr.AddAttr('autocopyrvf', '1')
  else
    Node.Attr.AddAttr('autocopyrvf', '0');

  if rvoAutoCopyImage in RichView.Options then
    Node.Attr.AddAttr('autocopyimage', '1')
  else
    Node.Attr.AddAttr('autocopyimage', '0');

  if rvoFormatInvalidate in RichView.Options then
    Node.Attr.AddAttr('formatinvalidate', '1')
  else
    Node.Attr.AddAttr('formatinvalidate', '0');

  if rvoDblClickSelectsWord in RichView.Options then
    Node.Attr.AddAttr('dblclickselectword', '1')
  else
    Node.Attr.AddAttr('dblclickselectword', '0');

  if rvoRClickDeselects in RichView.Options then
    Node.Attr.AddAttr('rightclickdeselect', '1')
  else
    Node.Attr.AddAttr('rightclickdeselect', '0');

  if RichView.CPEventKind = cpeNone then
    Node.Attr.AddAttr('checkevent', 'none')
  else if RichView.CPEventKind = cpeAsSectionStart then
    Node.Attr.AddAttr('checkevent', 'sectionstart')
  else if RichView.CPEventKind = cpeWhenVisible then
    Node.Attr.AddAttr('checkevent', 'visible');

  if RichView.BiDiMode = rvbdUnspecified then
    Node.Attr.AddAttr('bidimode', 'none')
  else if RichView.BiDiMode = rvbdLeftToRight then
    Node.Attr.AddAttr('bidimode', 'ltr')
  else if RichView.BiDiMode = rvbdRightToLeft then
    Node.Attr.AddAttr('bidimode', 'rtl');

  if rvrtfSaveStyleSheet in RichView.RTFOptions then
    Node.Attr.AddAttr('rtfsavestylesheet', '1')
  else
    Node.Attr.AddAttr('rtfsavestylesheet', '0');

  if rvrtfDuplicateUnicode in RichView.RTFOptions then
    Node.Attr.AddAttr('rtfunicodetoansi', '1')
  else
    Node.Attr.AddAttr('rtfunicodetoansi', '0');

  if rvrtfSaveEMFAsWMF in RichView.RTFOptions then
    Node.Attr.AddAttr('rtfemftowmf', '1')
  else
    Node.Attr.AddAttr('rtfemftowmf', '0');

  if rvrtfSaveJpegAsJpeg in RichView.RTFOptions then
    Node.Attr.AddAttr('rtftransformjpeg', '0')
  else
    Node.Attr.AddAttr('rtftransformjpeg', '1');

  if rvrtfSaveBitmapDefault in RichView.RTFOptions then
    Node.Attr.AddAttr('rtfdefaultimgformat', 'bmp')
  else if rvrtfSaveEMFDefault in RichView.RTFOptions then
    Node.Attr.AddAttr('rtfdefaultimgformat', 'emf')
  else
    Node.Attr.AddAttr('rtfdefaultimgformat', 'wmf');

  if rvfoSavePicturesBody in RichView.RVFOptions then
    Node.Attr.AddAttr('rvfautosaveimages', '1')
  else
    Node.Attr.AddAttr('rvfautosaveimages', '0');

  if rvfoSaveControlsBody in RichView.RVFOptions then
    Node.Attr.AddAttr('rvfautosavecontrols', '1')
  else
    Node.Attr.AddAttr('rvfautosavecontrols', '0');

  if rvfoUseStyleNames in RichView.RVFOptions then
    Node.Attr.AddAttr('rvfusestylenames', '1')
  else
    Node.Attr.AddAttr('rvfusestylenames', '0');

  if rvfoSaveTextStyles in RichView.RVFOptions then
    Node.Attr.AddAttr('rvfsavetextstyles', '1')
  else
    Node.Attr.AddAttr('rvfsavetextstyles', '0');

  if rvfoSaveParaStyles in RichView.RVFOptions then
    Node.Attr.AddAttr('rvfsaveparastyles', '1')
  else
    Node.Attr.AddAttr('rvfsaveparastyles', '0');

  if rvfoSaveBinary in RichView.RVFOptions then
    Node.Attr.AddAttr('rvfsavebinary', '1')
  else
    Node.Attr.AddAttr('rvfsavebinary', '0');

  if rvfoSaveBack in RichView.RVFOptions then
    Node.Attr.AddAttr('rvfsavebackground', '1')
  else
    Node.Attr.AddAttr('rvfsavebackground', '0');

  if rvfoSaveLayout in RichView.RVFOptions then
    Node.Attr.AddAttr('rvfsavelayout', '1')
  else
    Node.Attr.AddAttr('rvfsavelayout', '0');

  if not (RichView is TCustomRichViewEdit) then
  begin

    if RichView.TabNavigation = rvtnNone then
      Node.Attr.AddAttr('tabnavigation', 'none')
    else if RichView.TabNavigation = rvtnTab then
      Node.Attr.AddAttr('tabnavigation', 'tab')
    else if RichView.TabNavigation = rvtnCtrlTab then
      Node.Attr.AddAttr('tabnavigation', 'ctrltab');

    if rvoAllowSelection in RichView.Options then
      Node.Attr.AddAttr('allowselection', '1')
    else
      Node.Attr.AddAttr('allowselection', '0');

  end;

  if RichView is TCustomRichViewEdit then
  begin

    if rvoClearTagOnStyleApp in TCustomRichViewEdit(RichView).EditorOptions then
      Node.Attr.AddAttr('cleartagbystylechange', '1')
    else
      Node.Attr.AddAttr('cleartagbystylechange', '0');

    if rvoCtrlJumps in TCustomRichViewEdit(RichView).EditorOptions then
      Node.Attr.AddAttr('ctrljump', '1')
    else
      Node.Attr.AddAttr('ctrljump', '0');

    if rvoDoNotWantReturns in TCustomRichViewEdit(RichView).EditorOptions then
      Node.Attr.AddAttr('ignoreenter', '1')
    else
      Node.Attr.AddAttr('ignoreenter', '0');

    if rvoWantTabs in TCustomRichViewEdit(RichView).EditorOptions then
      Node.Attr.AddAttr('ignoretab', '0')
    else
      Node.Attr.AddAttr('ignoretab', '1');

    if rvoAutoSwitchLang in TCustomRichViewEdit(RichView).EditorOptions then
      Node.Attr.AddAttr('langautoswitch', '1')
    else
      Node.Attr.AddAttr('langautoswitch', '0');

  end;

end;

{------------------------------Save/Load RVPrint----------------------------}

procedure LoadRVPrint(Node: TXMLTag; RVPrint: TRVPrint);
var
  I: Integer;
  S: string;
begin
  for I := 0 to Node.Attr.Count - 1 do
  begin
    S := LowerCase(Node.Attr[I].Name);

    if S = 'bottommargin' then
      RVPrint.BottomMarginMM := StrToIntDef(Node.Attr[I].Value, RVPrint.BottomMarginMM)

    else if S = 'leftmargin' then
      RVPrint.LeftMarginMM := StrToIntDef(Node.Attr[I].Value, RVPrint.LeftMarginMM)

    else if S = 'rightmargin' then
      RVPrint.RightMarginMM := StrToIntDef(Node.Attr[I].Value, RVPrint.RightMarginMM)

    else if S = 'topmargin' then
      RVPrint.TopMarginMM := StrToIntDef(Node.Attr[I].Value, RVPrint.TopMarginMM)

    else if S = 'headermarginy' then
      RVPrint.HeaderYMM := StrToIntDef(Node.Attr[I].Value, RVPrint.HeaderYMM)

    else if S = 'footermarginy' then
      RVPrint.FooterYMM := StrToIntDef(Node.Attr[I].Value, RVPrint.FooterYMM)

    else if S = 'mirrormargins' then
    begin
      if Node.Attr[I].Value = '1' then
        RVPrint.MirrorMargins := True
      else if Node.Attr[I].Value = '0' then
        RVPrint.MirrorMargins := False;
    end

    else if S = 'previewcorrect' then
    begin
      if Node.Attr[I].Value = '1' then
        RVPrint.PreviewCorrection := True
      else if Node.Attr[I].Value = '0' then
        RVPrint.PreviewCorrection := False;
    end

    else if S = 'printback' then
    begin
      if Node.Attr[I].Value = '1' then
        RVPrint.TransparentBackground := False
      else if Node.Attr[I].Value = '0' then
        RVPrint.TransparentBackground := True;
    end;

  end;
end;

procedure SaveRVPrint(Node: TXMLTag; RVPrint: TRVPrint);
begin
  Node.Attr.AddAttr('bottommargin', IntToStr(RVPrint.BottomMarginMM));
  Node.Attr.AddAttr('leftmargin', IntToStr(RVPrint.LeftMarginMM));
  Node.Attr.AddAttr('rightmargin', IntToStr(RVPrint.RightMarginMM));
  Node.Attr.AddAttr('topmargin', IntToStr(RVPrint.TopMarginMM));
  Node.Attr.AddAttr('headermarginy', IntToStr(RVPrint.HeaderYMM));
  Node.Attr.AddAttr('footermarginy', IntToStr(RVPrint.FooterYMM));

  if RVPrint.MirrorMargins then
    Node.Attr.AddAttr('mirrormargins', '1')
  else
    Node.Attr.AddAttr('mirrormargins', '0');

  if RVPrint.PreviewCorrection then
    Node.Attr.AddAttr('previewcorrect', '1')
  else
    Node.Attr.AddAttr('previewcorrect', '0');

  if RVPrint.TransparentBackground then
    Node.Attr.AddAttr('printback', '0')
  else
    Node.Attr.AddAttr('printback', '1');
end;

{---------------------------------------------------------------------------}

function RVLoadFromXML(RichView: TCustomRichView; RVPrint: TRVPrint;
  LoadPrint: Boolean;
  Node: TXMLTag; var InsertPosition: Integer;
  LoadBackground: Boolean; StyleLoadingMode: TStyleLoadingMode;
  OnUnknownTag: TRVXMLUnknownTagEvent;
  OnLoadControl: TRVXMLLoadControlEvent;
  EditMode: Boolean; var FullReformat: Boolean;
  var NonFirstItemsAdded: Integer;
  LoadTagMode: Boolean): Boolean;
var
  Bullets: TImgListList;
  T: TXMLTag;
  A: TXMLAttr;
  StyleMap, ParaMap, ListMap: TRVIntegerList;
  FirstTime: Boolean;
begin
  NonFirstItemsAdded := 0;
  Result := False;
  StyleMap := TRVIntegerList.Create;
  ParaMap := TRVIntegerList.Create;
  ListMap := TRVIntegerList.Create;
  try
    T := Node.Items.FindTagOfName('style');
    LoadRVStyle(T, RichView.Style, StyleLoadingMode, StyleMap, ParaMap, ListMap);

    T := Node.Items.FindTagOfName('bullets');
    Bullets := TImgListList.Create;
    try
      if T <> nil then
        Bullets.LoadFromXML(T);

      if LoadPrint and (RVPrint <> nil) then
      begin
        T := Node.Items.FindTagOfName('print');
        if T <> nil then
          LoadRVPrint(T, RVPrint);
      end;

      T := Node.Items.FindTagOfName('document');
      if T = nil then begin
        Bullets.Free;
        Exit;
      end;

      if LoadTagMode then
      begin
        A := T.Attr.FindAttrOfName('stringtags');
        if A <> nil then
        begin
          if A.Value = '1' then
            RichView.Options := RichView.Options + [rvoTagsArePChars]
          else if A.Value = '0' then
            RichView.Options := RichView.Options - [rvoTagsArePChars];
        end;
      end;
      if LoadBackground then
        LoadBackgroundFromXML(T, RichView);

      FirstTime := True;
      FullReformat := False;
      Result := LoadRVDataFromXML(RichView.RVData, T, Bullets, nil, InsertPosition, RichView.DoInPaletteMode, RichView.RVPalette,
        RichView.PRVLogPalette, nil, StyleMap, ParaMap, ListMap, OnUnknownTag, OnLoadControl, RichView.Style, EditMode, FirstTime, FullReformat,
         NonFirstItemsAdded, rvoTagsArePChars in RichView.Options);

      if not EditMode then
        RichView.Format;
    finally
      Bullets.Free;
    end;
  finally
    StyleMap.Free;
    ParaMap.Free;
    ListMap.Free;
  end;
end;

procedure RVSaveToXML(RichView: TCustomRichView; Node: TXMLTag;
  RVPrint: TRVPrint; SavePrint: Boolean;
  Style: TRichViewXMLStyle; SaveStyles: Boolean;
  SaveBackground: Boolean; SaveImgToFiles: Boolean;
  SelStart, SelOffset, SelEnd, SelEndOffset: Integer;
  OnUnknownItem: TRVXMLUnknownItemEvent;
  OnSaveControl: TRVXMLSaveControlEvent;
  UseStyleNames: Boolean; SaveStyleNames: Boolean);
var
  Bullets: TImgListList;
  T: TXMLTag;
begin
  if SaveStyles then
  begin
    T := Node.Items.AddTag('style', Node, Node.Root);
    SaveRVStyle(T, RichView.Style, SaveStyleNames, SaveImgToFiles);
  end;

  Bullets := TImgListList.Create;
  try
    ScanBullets(RichView.RVData, nil, Bullets, SelStart, SelEnd);
    if Bullets.Count > 0 then
    begin
      T := Node.Items.AddTag('bullets', Node, Node.Root);
      Bullets.SaveToXML(T, SaveImgToFiles);
    end;

    if SavePrint and (RVPrint <> nil) then
    begin
      T := Node.Items.AddTag('print', Node, Node.Root);
      SaveRVPrint(T, RVPrint);
    end;

    T := Node.Items.AddTag('document', Node, Node.Root);

    if rvoTagsArePChars in RichView.Options then
      T.Attr.AddAttr('stringtags', '1')
    else
      T.Attr.AddAttr('stringtags', '0');

    if SaveBackground then SaveBackgroundToXML(T, RichView, SaveImgToFiles);
    SaveRVDataToXML(RichView.RVData, T, Style, True, RichView.Style, SaveImgToFiles, Bullets,
      SelStart, SelOffset, SelEnd, SelEndOffset, rvoTagsArePChars in RichView.Options, OnUnknownItem,
      OnSaveControl, UseStyleNames)
  finally
    Bullets.Free;
  end;
end;

end.
